#
# Copyright 2018, 2020-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Functions for regression of noisy data and resampling/averaging from one grid
to another. Currently implemented are simple bin averages and Gaussian process
regression.
"""

import logging

import numpy as np

from ..Exceptions import NoReliableDataError

_log = logging.Logger(__name__)


def make_grid(collocation, min_value, max_value, nb_points=None, nb_points_per_decade=10, dectol=0.01):
    """
    Create collocation points.

    Parameters
    ----------
    collocation : {'log', 'quadratic', 'linear', array_like}
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin edges are explicitly specified, then the other arguments
        to this function are ignored.
    min_value : float
        Minimum value. Note that for log-spaced collocation points, this
        is the first value - the leftmost bin edge with always be zero.
    max_value : float
        Maximum value.
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)
    dectol : float, optional
        Tolerance for left and right bin edge in terms of decades added to
        the edges. (Default: 0.01)

    Returns
    -------
    collocation_points : np.ndarray
        List of collocation points.
    bin_edges : np.ndarray
        List of bins edges; collocations points are located within the
        respective bin. This array contains one more data point than
        the `collocation_points` array.
    """
    if collocation == 'log':
        # Power law -> equally spaced on a log-log plot
        if nb_points is None:
            # The size of the first bin should be equal to min_radius,
            # which is the (minimal) size of a grid point
            # i.e. min_radius = np.exp(np.log(min_radius) + dl) - min_radius
            # => dl = np.log(2)
            nb_points = int(np.ceil((np.log10(max_value) - np.log10(min_value) + 1) * nb_points_per_decade))
        bin_edges = np.linspace(np.log10(min_value) - dectol, np.log10(max_value) + dectol, nb_points)
        collocation_points = 10 ** ((bin_edges[1:] + bin_edges[:-1]) / 2)
        bin_edges = 10 ** bin_edges
    elif collocation == 'quadratic':
        # Quadratic -> similar statistics for each data point on a 2D radial grid
        if nb_points is None:
            raise ValueError("Please specify number of bins for 'quadratic' bins.")
        bin_edges = np.sqrt(np.linspace(min_value ** 2, max_value ** 2, nb_points + 1))
        collocation_points = (bin_edges[1:] + bin_edges[:-1]) / 2
    elif collocation == 'linear':
        # Linear
        if nb_points is None:
            raise ValueError("Please specify number of bins for 'linear' bins.")
        bin_edges = np.linspace(min_value, max_value, nb_points + 1)
        collocation_points = (bin_edges[1:] + bin_edges[:-1]) / 2
    else:
        bin_edges = collocation
        collocation_points = (bin_edges[1:] + bin_edges[:-1]) / 2
    return collocation_points, bin_edges


def gaussian_kernel(x1, x2, length_scale=1, signal_variance=1):
    return signal_variance * np.exp(-(x1 - x2) ** 2 / (2 * length_scale ** 2))


def gaussian_log_kernel(x1, x2, length_scale=1, signal_variance=1):
    return signal_variance * np.exp(-(np.log10(x1) - np.log10(x2)) ** 2 / (2 * length_scale ** 2))


def suggest_kernel_for_grid(collocation, nb_collocation_points, min_value, max_value):
    """
    Suggest a kernel function for Gaussian process regression with test
    locations on a specific grid. The length scale (smoothening scale)
    of the kernel is set to the characteristic spacing between the sampling
    output points.

    Parameters
    ----------
    collocation : {'log', 'quadratic', 'linear', array_like}
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin edges are explicitly specified, then the other arguments
        to this function are ignored.
    nb_collocation_points : int
        Number of collocation points.
    min_value : float
        Minimum value. Note that for log-spaced collocation points, this
        is the first value - the leftmost bin edge with always be zero.
    max_value : float
        Maximum value.

    Returns
    -------
    kernel : func
        Kernel function.
    """
    if collocation == 'log':
        length_scale = (np.log10(max_value) - np.log10(min_value)) / nb_collocation_points
        return lambda x1, x2: gaussian_log_kernel(x1, x2, length_scale=length_scale)
    else:
        length_scale = (max_value - min_value) / nb_collocation_points
        return lambda x1, x2: gaussian_kernel(x1, x2, length_scale=length_scale)


def bin_average(collocation_points, bin_edges, x, values):
    """
    Average values over bins. Returns NaN for bins without data.

    Parameters
    ----------
    collocation_points : np.ndarray
        Collocation points.
    bin_edges : array_like
        Edges of bins.
    x : array_like
        Collocation points for `values` array.
    values : array_like
        Function values/variates.

    Returns
    -------
    resampled_collocation_points : np.ndarray
        Collocation points, resampled as average over input `x` if data is present.
    resampled_values : np.ndarray
        Resampled/averaged values.
    resampled_variance : np.ndarray
        Variance of resampled data.
    """
    # Find bin index
    bin_index = np.searchsorted(bin_edges, x)

    # Compute averages within bins
    number_of_data_points = np.bincount(bin_index, minlength=len(bin_edges) + 1)
    number_of_data_points1 = np.where(number_of_data_points == 0, np.ones_like(number_of_data_points),
                                      number_of_data_points)
    resampled_values = np.bincount(bin_index, weights=values, minlength=len(bin_edges) + 1) / number_of_data_points1
    resampled_variance = \
        np.bincount(bin_index, weights=values ** 2,
                    minlength=len(bin_edges) + 1) / number_of_data_points1 - resampled_values ** 2

    # Resample collocation points as average of the distances in each bin
    resampled_collocation_points = np.bincount(bin_index, weights=x,
                                               minlength=len(bin_edges) + 1) / number_of_data_points1

    # We discard the final element as it contains data points outside our binned region
    resampled_collocation_points = resampled_collocation_points[1:-1]
    resampled_values = resampled_values[1:-1]
    resampled_variance = resampled_variance[1:-1]

    # Mark elements with no data
    mask = number_of_data_points[1:-1] == 0
    resampled_collocation_points[mask] = collocation_points[mask]
    resampled_values[mask] = np.nan
    resampled_variance[mask] = np.nan
    return resampled_collocation_points, resampled_values, resampled_variance


def gaussian_process_regression(output_x, x, values, kernel=gaussian_kernel, noise_variance=1e-6):
    """
    Gaussian process regression for resampling a simple function.

    Parameters
    ----------
    output_x : array_like
        Collocation points for resampled values.
    x : array_like
        Collocation points for `values` array.
    values : array_like
        Function values/variates.
    kernel : func, optional
        Kernel function/covariance model.
        (Default: gaussian_kernel)
    noise_variance : float, optional
        Noise variance.
        (Default: 1e-6)

    Returns
    -------
    resampled_values : np.ndarray
        Resampled/averaged values.
    """
    # Covariance between observations
    obs_cov = kernel(x.reshape(-1, 1), x.reshape(1, -1))

    # Add noise to observation covariance matrix
    obs_cov += noise_variance * np.identity(len(x))

    # Compute kernel coefficients
    coeff = np.linalg.solve(obs_cov, values)

    # Covariance between test outputs
    test_cov = kernel(output_x.reshape(-1, 1), output_x.reshape(1, -1))

    # Covariance between observation and test outputs
    obs_test_cov = kernel(x.reshape(-1, 1), output_x.reshape(1, -1))

    # Compute predictive mean
    pred_mean = coeff.dot(obs_test_cov)

    # Compute predictive covariance
    pred_cov = test_cov - obs_test_cov.T.dot(np.linalg.solve(obs_cov, obs_test_cov))

    # Return mean and variance
    return pred_mean, pred_cov.diagonal()


def resample(x, values, collocation='log', nb_points=None, min_value=None, max_value=None, nb_points_per_decade=10,
             method='bin-average'):
    """
    Resample noisy function data set onto a specific grid.

    Parameters
    ----------
    x : array_like
        Evaluation points.
    values : array_like
        Function values.
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    min_value : float, optional
        Minimum value, is automatically determined from the range of the data
        if not provided. (Default: None)
    max_value : float, optional
        Maximum value, is automatically determined from the range of the data
        if not provided. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)
    method : str, optional
        Method can be 'bin-average' for simple bin averaging and
        'gaussian-process' for Gaussian process regression.
        (Default: 'bin-average')

    Returns
    -------
    collocation_points : np.ndarray
        Points where the data has been collected.
    bin_edges : np.ndarray
        Bin edges.
    resampled_values : np.ndarray
        Resampled values.
    resampled_variance : np.ndarray
        Variance of resampled data.
    """
    # pylint: disable=invalid-name
    if max_value is None:
        max_value = np.max(x)
    if min_value is None:
        min_value = np.min(x)

    if min_value >= max_value:
        raise NoReliableDataError('Dataset contains no reliable data.')

    collocation_points, bin_edges = make_grid(collocation, min_value, max_value, nb_points=nb_points,
                                              nb_points_per_decade=nb_points_per_decade)

    if method == 'bin-average':
        collocation_points, resampled_values, resampled_variance = bin_average(collocation_points, bin_edges, x, values)
    elif method == 'gaussian-process':
        if collocation == 'log':
            # For log-spaced sampling points we fit in log-space... because
            # this is then most likely a power-law. Note that this is pure
            # function fitting - if we are collecting noisy data to compute
            # the ACF or PSD we should actually be averaging in linear space.
            # (This is because the arithmetic mean is the max likelihood
            # estimator for the underlying distribution of the data. The
            # difference can be clearly seen in numerical experiments.)
            resampled_values, resampled_variance = \
                gaussian_process_regression(np.log(collocation_points), np.log(x), np.log(values))
            resampled_values = np.exp(resampled_values)
        else:
            _log.warning('Gaussian process regression of log-spaced data should only be used for fitting functions, '
                         'not for inference from noisy data.')
            resampled_values, resampled_variance = \
                gaussian_process_regression(collocation_points, x, values)
    # FIXME: The variance is bogus when log-sampling and we should really use a different kernel function in that case.
    #        resampled_values, resampled_variance = \
    #            gaussian_process_regression(collocation_points, x, values,
    #                                        kernel=suggest_kernel_for_grid(collocation, len(collocation_points),
    #                                                                       min_value, max_value))
    else:
        raise ValueError(f"Unknown resampling method '{method}'.")

    return collocation_points, bin_edges, resampled_values, resampled_variance


def resample_radial(data, physical_sizes=None, collocation='log', nb_points=None, min_radius=0, max_radius=None,
                    nb_points_per_decade=5, full=True, method='bin-average'):
    """
    Compute radial average of quantities reported on a 2D grid and collect
    results of collocation points.

    Parameters
    ----------
    data : array_like
        2D-array of values to be averaged.
    physical_sizes : (float, float), optional
        Physical size of the 2D grid. (Default: Size is equal to number of
        grid points.)
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    min_radius : float, optional
        Minimum radius. (Default: 0)
    max_radius : float, optional
        Maximum radius, is automatically determined from the range of the data
        if not provided. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)
    full : bool, optional
        Number of quadrants contained in data.
        True: Full radial average from 0 to 2*pi.
        False: Only the one quarter of the full circle is present. Radial
        average from 0 to pi/2.
        (Default: True)
    method : str, optional
        Method can be 'bin-average' for simple bin averaging and
        'gaussian-process' for Gaussian process regression.
        (Default: 'bin-average')

    Returns
    -------
    collocation_points : np.ndarray
        Points where the data has been collected.
    bin_edges : np.ndarray
        Bin edges.
    resampled_values : np.ndarray
        Resampled values.
    resampled_variance : np.ndarray
        Variance of resampled data.
    """
    # pylint: disable=invalid-name
    nx, ny = data.shape
    x = np.arange(nx)
    y = np.arange(ny)
    if full:
        x = np.where(x > nx // 2, nx - x, x)
        y = np.where(y > ny // 2, ny - y, y)

    if physical_sizes is None:
        min_radius = max(min_radius, 1.0)
    else:
        sx, sy = physical_sizes
        x = sx / nx * x
        y = sy / ny * y
        min_radius = max(min(sx / nx, sy / ny), min_radius)
    radius = np.sqrt((x ** 2).reshape(-1, 1) + (y ** 2).reshape(1, -1))
    if max_radius is None:
        max_radius = np.max(radius)

    return resample(np.ravel(radius), np.ravel(data), collocation=collocation, nb_points=nb_points,
                    min_value=min_radius, max_value=max_radius, nb_points_per_decade=nb_points_per_decade,
                    method=method)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

from SurfaceTopography.Support.Interpolation import Bicubic

nx = ny = 8192
tol = 1e-9

NPARA = 4 * 4
print(f"Estimated memory requirement: {nx * ny * np.dtype('f8').itemsize * (NPARA + 1) / 1024. / 1024. / 1024.} GB")

field = np.random.random([nx, ny])
interp = Bicubic(field)
for i in range(nx):
    for j in range(ny):
        assert abs(interp(i, j) - field[i, j]) < tol

x, y = np.mgrid[:nx, :ny]

for der in [0, 1, 2]:
    if der == 0:
        interp_field = interp(x, y, derivative=der)
    elif der == 1:
        interp_field, _, _ = interp(x, y, derivative=der)
    else:
        interp_field, _, _, _, _, _ = interp(x, y, derivative=der)
    assert np.allclose(interp_field, field)



#
# Copyright 2022 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#
# Reference information and implementations:
# https://sourceforge.net/p/gwyddion/code/HEAD/tree/trunk/gwyddion/modules/file/surffile.c
#

import datetime

import numpy as np

from .binary import decode
from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo
from ..Exceptions import CorruptFile, FileFormatMismatch, MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography
from ..Support.UnitConversion import get_unit_conversion_factor


class SURReader(ReaderBase):
    _format = 'sur'
    _name = 'Digital Surf'
    _description = '''
This reader imports Digital Surf SUR data files.
'''

    _MAGIC = b'DIGITAL SURF'

    _header_structure = [
        ('format', 'H'),
        ('nb_objects', 'H'),
        ('version', 'H'),
        ('type', 'H'),
        ('object_name', '30s'),
        ('instrument_name', '30s'),
        ('material_code', 'H'),
        ('acquisition', 'H'),
        ('range', 'H'),
        ('special_points', 'H'),
        ('absolute', 'H'),
        ('_', 'I'),
        ('_', 'I'),
        ('itemsize', 'H'),
        ('zmin', 'i'),
        ('zmax', 'i'),
        ('nb_grid_pts_x', 'i'),
        ('nb_grid_pts_y', 'i'),
        ('nb_points', 'I'),
        ('grid_spacing_x', 'f'),
        ('grid_spacing_y', 'f'),
        ('height_scale_factor', 'f'),
        ('name_x', '16s'),
        ('name_y', '16s'),
        ('data_name', '16s'),
        ('unit_delta_x', '16s'),
        ('unit_delta_y', '16s'),
        ('delta_data_unit', '16s'),
        ('unit_x', '16s'),
        ('unit_y', '16s'),
        ('data_unit', '16s'),
        ('unit_ratio_x', 'f'),
        ('unit_ratio_y', 'f'),
        ('data_unit_ratio', 'f'),
        ('imprint', 'H'),
        ('inversion', 'H'),
        ('leveling', 'H'),
        ('_', 'I'),
        ('_', 'I'),
        ('_', 'I'),
        ('second', 'H'),
        ('minute', 'H'),
        ('hour', 'H'),
        ('day', 'H'),
        ('month', 'H'),
        ('year', 'H'),
        ('dayof', 'H'),
        ('measurement_duration', 'f'),
        ('_', 'I'),
        ('_', 'I'),
        ('_', 'H'),
        ('comment_size', 'H'),
        ('private_size', 'H'),
        ('client_zone', '128s'),
        ('x_offset', 'f'),
        ('y_offset', 'f'),
        ('data_offset', 'f'),
    ]

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        self.file_path = file_path
        with OpenFromAny(file_path, 'rb') as f:
            # Detect file magic
            if f.read(len(self._MAGIC)) != self._MAGIC:
                raise FileFormatMismatch('File magic does not match. This is not a Digital Surf file.')

            self._header = decode(f, self._header_structure, '<')

        nx = self._header['nb_grid_pts_x']
        ny = self._header['nb_grid_pts_y']
        if nx * ny != self._header['nb_points']:
            raise CorruptFile(
                'The file reported a grid of {} x {} data points and a total number of {} data points, which is '
                'inconsistent'.format(self._nb_grid_pts[0], self._nb_grid_pts[1], self._header['nb_points']))

        if self._header['itemsize'] not in [16, 32]:
            raise CorruptFile('The file reported an item size of {} bits, which I cannot read.'
                              .format(self._header['itemsize']))

        # Check that units and delta units are the same. Not sure if they differ and why there are two different sets
        # of units provided.
        unit_x = self._header['unit_x']
        unit_y = self._header['unit_y']
        self._unit = self._header['data_unit']  # We use the data unit as the primary unit for topography objects

        if self._header['unit_delta_x'] != unit_x or self._header['unit_delta_y'] != unit_y or \
                self._header['delta_data_unit'] != self._unit:
            raise CorruptFile('Units and delta units differ. Not sure how to handle this.')

        # Get the conversion factors for converting x,y to the main system of units
        fac_x = get_unit_conversion_factor(unit_x, self._unit)
        fac_y = get_unit_conversion_factor(unit_y, self._unit)

        # All good, now initialize some convenience variables

        self._nb_grid_pts = (nx, ny)
        self._physical_sizes = \
            (fac_x * self._header['grid_spacing_x'] * nx, fac_y * self._header['grid_spacing_y'] * ny)

        self._info = {
            'acquisition_time':
                str(datetime.datetime(self._header['year'], self._header['month'], self._header['day'],
                                      self._header['hour'], self._header['minute'], self._header['second'])),
            'instrument': {'name': self._header['instrument_name']},
            'raw_metadata': self._header
        }

    def read_height_data(self, f):
        if self._header['itemsize'] == 16:
            dtype = np.dtype('<i2')
        elif self._header['itemsize'] == 32:
            dtype = np.dtype('<i4')
        else:
            raise RuntimeError('Unknown itemsize')  # Should not happen because we check this in the constructor

        f.seek(512)

        nx, ny = self._nb_grid_pts
        buffer = f.read(nx * ny * np.dtype(dtype).itemsize)
        return np.frombuffer(buffer, dtype=dtype).reshape(self._nb_grid_pts)

    @property
    def channels(self):
        return [ChannelInfo(self,
                            0,  # channel index
                            name='Default',
                            dim=2,
                            nb_grid_pts=self._nb_grid_pts,
                            physical_sizes=self._physical_sizes,
                            uniform=True,
                            unit=self._unit)]

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={},
                   periodic=None, subdomain_locations=None,
                   nb_subdomain_grid_pts=None):
        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError('This reader does not support MPI parallelization.')

        if channel_index is None:
            channel_index = self._default_channel_index

        if channel_index != self._default_channel_index:
            raise RuntimeError(f'There is only a single channel. Channel index must be {self._default_channel_index}.')

        if physical_sizes is not None:
            raise MetadataAlreadyFixedByFile('physical_sizes')

        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')

        with OpenFromAny(self.file_path, 'rb') as f:
            height_data = self.read_height_data(f)

        _info = self._info.copy()
        _info.update(info)

        topo = Topography(height_data,
                          self._physical_sizes,
                          unit=self._unit,
                          periodic=False if periodic is None else periodic,
                          info=_info)
        return topo.scale(float(self._header['height_scale_factor']))



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import io
import requests
import tempfile
import textwrap
import yaml
import numpy as np
from datetime import datetime
from zipfile import ZipFile

try:
    from importlib.metadata import version

    __version__ = version('SurfaceTopography')
except ImportError:
    from pkg_resources import get_distribution

    __version__ = get_distribution('SurfaceTopography').version

from ..IO import open_topography
from .SurfaceContainer import SurfaceContainer


def read_container(fn, datafile_keys=['original', 'squeezed-netcdf']):
    """
    Read all surface in a container file and associated metadata. The
    container is a ZIP file with raw data files and a YAML file meta.yml
    that contains all metadata not contained in the raw data files.

    Parameters
    ----------
    fn : str or stream
        File or stream that contains the ZIP-container.
    datafile_keys : list of str, optional
        List of possible keys in 'meta.yml' that contains the name of the
        datafile to open. Code will try these keys in order. If a key
        starts with 'squeezed', the pipeline is not constructed from
        the metadata.
        (Default: ['original', 'squeezed-netcdf'])

    Returns
    -------
    surface_containers : list of :obj:`SurfaceContainer`s
        List of all surfaces contained in this container file.
    """

    surfaces = []

    with ZipFile(fn, 'r') as z:
        meta = yaml.load(z.open('meta.yml'), Loader=yaml.FullLoader)

        topographies = []
        for surf_meta in meta['surfaces']:
            for topo_meta in surf_meta['topographies']:
                info = topo_meta.copy()

                # Check whether the metadata contains sizes and unit. If not
                # this information comes from the data file.
                try:
                    physical_sizes = info['size']
                    del info['size']
                except KeyError:
                    physical_sizes = None
                try:
                    unit = info['unit']
                    del info['unit']
                except KeyError:
                    unit = None
                try:
                    periodic = info['is_periodic']
                    del info['is_periodic']
                except KeyError:
                    periodic = False
                datafile_key = None
                datafiles = topo_meta['datafile']

                # Pick first of the provided possible data file keys that
                # exists in the container
                for key in datafile_keys:
                    if key in datafiles:
                        datafile_key = key
                        break

                # There may be none, complain
                if datafile_key is None:
                    raise ValueError('Could not detect data file.')

                # Inspect topography file
                r = open_topography(z.open(datafiles[datafile_key]))

                # Channel to load
                if 'data_source' in topo_meta:
                    data_source = topo_meta['data_source']
                else:
                    data_source = r.default_channel.index

                # Check consistency between data file and meta.yml
                physical_sizes_from_file = r.channels[data_source].physical_sizes
                if physical_sizes_from_file is not None:
                    if physical_sizes is not None:
                        if np.allclose(physical_sizes_from_file, physical_sizes, rtol=1e-4):
                            # Need to set this to None to avoid collision
                            physical_sizes = None
                        else:
                            raise ValueError(f'Physical sizes from data file (={physical_sizes_from_file} and from '
                                             f'meta.yml (={physical_sizes}) differ for topography '
                                             f'{datafiles[datafile_key]}')

                unit_from_file = r.channels[data_source].unit
                if unit_from_file is not None:
                    if unit is not None:
                        if unit_from_file == unit:
                            # Need to set this to None to avoid collision
                            unit = None
                        else:
                            raise ValueError(f'Unit from data file (={unit_from_file}) and from meta.yml '
                                             f'(={unit}) differ for topography {datafiles[datafile_key]}')

                # Read the topography from the preferred data file
                t = r.topography(
                    physical_sizes=physical_sizes,
                    periodic=periodic,
                    unit=unit,
                    info=info
                )

                # Close reader
                r.close()

                # We need to reconstruct the pipeline if the data file does
                # not contain squeezed data, currently indicate by a
                # 'squeezed' prefix to the data file key
                if not datafile_key.startswith('squeezed'):
                    if 'height_scale' in topo_meta:
                        t = t.scale(topo_meta['height_scale'])
                    if 'detrend_mode' in topo_meta:
                        t = t.detrend(topo_meta['detrend_mode'])
                topographies += [t]

            surfaces += [SurfaceContainer(topographies)]

    return surfaces


def read_published_container(publication_url):
    # First get the page for the publication in order
    # to get the download URL
    publication_response = requests.get(publication_url)
    download_url = publication_response.url + "download/"

    # Then download and read container
    container_response = requests.get(download_url)
    container_file = io.BytesIO(container_response.content)
    return read_container(container_file)


def write_containers(containers, fn):
    """
    Write multiple surface containers into a ZIP file.

    Parameters
    ----------
    containers : list of :obj:`SurfaceContainer`s
        Containers to be written to the ZIP file.
    fn : str or stream
        File or stream to write the ZIP-container to.
    """

    # This is adapted from TopoBank

    surfaces_dicts = []
    counter = 0

    with ZipFile(fn, mode='w') as zf:
        #
        # Add meta data and topography files for all given surfaces
        #
        for surface_container in containers:
            topography_dicts = []

            for topography in surface_container:
                # Create unique file names for the data files by simply appending a counter
                topofile_name = f'topography{counter}.nc'
                counter += 1

                # We do not write the info dictionary into the YAML metadata as it is already container in the NetCDF
                # file (serialized as JSON)
                topo_meta = {
                    'datafile': {
                        'squeezed-netcdf': topofile_name
                    }
                }

                # Add topography file as NetCDF to the ZIP archive
                with tempfile.TemporaryFile() as f:
                    topography.to_netcdf(f)
                    f.seek(0)
                    zf.writestr(topofile_name, f.read())

                topography_dicts.append(topo_meta)

            surface_dict = {'topographies': topography_dicts}
            surfaces_dicts.append(surface_dict)

        #
        # Add metadata file
        #
        metadata = {
            'versions': {'SurfaceTopography': __version__},
            'surfaces': surfaces_dicts,
            'creation_time': str(datetime.now()),
        }

        zf.writestr("meta.yml", yaml.dump(metadata))

        #
        # Add a Readme file
        #
        readme_txt = textwrap.dedent(f"""
        Contents of this ZIP archive
        ============================
        This archive contains {len(containers)} surface(s). Each surface is a
        collection of individual topography measurements.
        In total {sum(len(x) for x in containers)} topography measurements are included.

        The meta data for the surfaces and the individual topographies
        can be found in the auxiliary file 'meta.yml' and within
        the NetCDF topographies themselves. 'meta.yml' is formatted
        as a [YAML](https://yaml.org/) file. The NetCDF topographies
        contain metadata as [JSON](https://www.json.org/).

        Version information
        ===================

        SurfaceTopography: {__version__}
        """)

        zf.writestr("README.txt", textwrap.dedent(readme_txt))


SurfaceContainer.register_function('to_zip', lambda container, fn: write_containers([container], fn))



#
# Copyright 2019-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Utility function for conversion between uniform and nonuniform
representations.
"""

import numpy as np

from SurfaceTopography.HeightContainer import UniformTopographyInterface
from SurfaceTopography.NonuniformLineScan import NonuniformLineScan, DecoratedNonuniformTopography


class WrapAsNonuniformLineScan(DecoratedNonuniformTopography):
    """
    Wrap a uniform topography into a nonuniform one.
    """

    def __init__(self, topography, info={}):
        """
        Parameters
        ----------
        topography : :obj:`NonuniformLineScan`
            SurfaceTopography to wrap.
        info : dict
            Additional entries for the info dictionary.
        """
        super().__init__(topography, info=info)

        if topography.dim != 1:
            raise ValueError('Only (one-dimensional) uniform line scans can be turned into nonuniform line scans.')

        if topography.is_periodic:
            raise ValueError('Periodic uniform line scans cannot be turned into nonuniform line scans.')

        # This is populated with functions from the nonuniform topography, but
        # this is a uniform topography
        self._functions = NonuniformLineScan._functions

    # Implement abstract methods of AbstractHeightContainer

    @property
    def dim(self):
        return 1

    @property
    def physical_sizes(self):
        s, = self.parent_topography.physical_sizes
        p, = self.parent_topography.pixel_size
        return s - p,

    @property
    def is_periodic(self):
        return False

    # Implement nonuniform line scan interface

    @property
    def nb_grid_pts(self):
        return self.parent_topography.nb_grid_pts

    @property
    def x_range(self):
        s, = self.parent_topography.physical_sizes
        p, = self.parent_topography.pixel_size
        return 0, s - p

    def positions(self):
        """
        Returns array containing the lateral positions.
        """
        r, = self.parent_topography.nb_grid_pts
        p, = self.parent_topography.pixel_size
        return np.arange(r) * p

    def heights(self):
        """
        Returns array containing the topography data.
        """
        return self.parent_topography.heights()


# Register pipeline functions from this module
UniformTopographyInterface.register_function('to_nonuniform', WrapAsNonuniformLineScan)



#
# Copyright 2019-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Height-difference autocorrelation functions for nonuniform line scans
"""

import numpy as np

import _SurfaceTopography

from ..Exceptions import ReentrantDataError, NoReliableDataError
from ..HeightContainer import NonuniformLineScanInterface


def height_height_autocorrelation(self, distances=None):
    r"""
    Compute the one-dimensional height-height autocorrelation function
    (ACF).

    This function treats the nonuniform line scan as a piece-wise function of
    straight lines between the data points. The ACF is computed exactly for
    this piece-wise linear interpolation of the data.

    Parameters
    ----------
    self : :obj:`NonuniformLineScan`
        Container storing the nonuniform line scan.
    r : array_like
        Array containing distances for which to compute the ACF. If no array
        is given, the function will automatically construct an array with
        equally spaced distances. (Default: None)

    Returns
    -------
    distances : array
        Distances. (Units: length)
    A : array
        Autocorrelation function. (Units: length**2)
    """
    if self.is_reentrant:
        raise ReentrantDataError('This topography is reentrant (i.e. it contains overhangs). The autocorrelation '
                                 'cannot be computed for reentrant topographies.')

    size, = self.physical_sizes
    if distances is None:
        # FIXME!!! We need a better heuristics to decide on the distances
        res, = self.nb_grid_pts
        distances = np.linspace(0, size, res)
    else:
        distances = np.asarray(distances, dtype=float)
    A = np.zeros_like(distances)

    x, h = self.positions_and_heights()
    s = self.derivative(1)
    # FIXME!!! This is slow
    for i in range(len(x) - 1):
        for j in range(len(x) - 1):
            # Determine lower and upper distance between segment i, i+1 and
            # segment j, j+1
            x1 = x[i]
            x2 = x[j]
            h1 = h[i]
            h2 = h[j]
            s1 = s[i]
            s2 = s[j]
            b1 = np.maximum(x1, x2 - distances)
            b2 = np.minimum(x[i + 1], x[j + 1] - distances)
            b = (b1 + b2) / 2
            db = (b2 - b1) / 2
            m = db > 0
            if m.sum() > 0:
                b = b[m]
                db = db[m]
                # f1[x_] := (h1 + s1*(x - x1))
                # f2[x_] := (h2 + s2*(x - x2))
                # FullSimplify[Integrate[f1[x]*f2[x + d],
                # {x, b - db, b + db}]]
                #   = 2 * f1[b] * f2[b + d] * db + 2 * s1 * s2 * db ** 3 / 3
                A[m] += 2 * (h1 + s1 * (b - x1)) * (
                            h2 + s2 * (b + distances[m] - x2)) * db + 2 * (
                                s1 * s2 * db ** 3) / 3
    return distances, A


def height_difference_autocorrelation(self, reliable=True, algorithm='fft', distances=None, nb_interpolate=5,
                                      short_cutoff=np.mean, resampling_method='bin-average', collocation='log',
                                      nb_points=None, nb_points_per_decade=10):
    r"""
    Compute the one-dimensional height-difference autocorrelation function
    (ACF).

    This function treats the nonuniform line scan as a piece-wise function of
    straight lines between the data points. The ACF is computed exactly for
    this piece-wise linear interpolation of the data.

    Parameters
    ----------
    self : :obj:`NonuniformLineScan`
        Container storing the nonuniform line scan.
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    algorithm : str
        Algorithm to compute autocorrelation.
        * 'fft': Interpolates the nonuniform line scan on a grid and then uses
        the FFT to compute the autocorrelation. Scales O(N log N)
        * 'brute-force': Brute-force computation using between line segements.
        Scale O(N^2 M) where M is the number of distance point.
        (Default: 'fft')
    distances : array_like
        Array containing distances for which to compute the ACF. If no array
        is given, the function will automatically construct an array with
        equally spaced distances. Can be used only if 'brute-force'
        algorithm is used. (Default: None)
    nb_interpolate : int
        Number of grid points to put between closest points on surface. Only
        used for 'fft' algorithm. (Default: 5)
    short_cutoff : function, optional
        Function that determines how the short cutoff for the returned PSD is
        computed. If set to None, the full PSD of the interpolated data is
        returned. If the user passes a function, that function is called with
        an array that contains the distances between the points of the uniform
        line scan. Pass `np.max`, `np.mean` or `np.min` to use the maximum,
        mean or minimum of that distance as the cutoff.
    resampling_method : str, optional
        Method can be None for no resampling (return on the grid of the
        data) 'bin-average' for simple bin averaging and 'gaussian-process'
        for Gaussian process regression.
        (Default: None)
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)

    Returns
    -------
    distances : array
        Distances. (Units: length)
    acf : array
        Autocorrelation function. (Units: length**2)
    """
    if self.is_reentrant:
        raise ReentrantDataError('This topography is reentrant (i.e. it contains overhangs). The autocorrelation '
                                 'cannot be computed for reentrant topographies.')

    s, = self.physical_sizes
    x, h = self.positions_and_heights()
    if algorithm == 'fft':
        if distances is not None:
            raise ValueError("`distances` can only be used with 'brute-force' algorithm.")
        distances, acf = self.to_uniform(nb_interpolate=nb_interpolate) \
            .autocorrelation_from_profile(reliable=reliable, resampling_method=resampling_method,
                                          collocation=collocation, nb_points=nb_points,
                                          nb_points_per_decade=nb_points_per_decade)
    elif algorithm == 'brute-force':
        if reliable:
            raise ValueError("'Brute-force' algorithm for nonuniform line scans does not support reliability analysis.")
        if resampling_method is not None:
            raise ValueError("'Brute-force' algorithm for nonuniform line scans does not support resampling.")
        if distances is None:
            raise ValueError("You need to specify `distances` for the 'brute-force' algorithm.")
        distances, acf = _SurfaceTopography.nonuniform_autocorrelation(x, h, s, distances)
    else:
        raise ValueError("Unknown algorithm '{}' specified.".format(algorithm))

    if short_cutoff is not None:
        mask = distances > short_cutoff(np.diff(x)) / 2
        if mask.sum() == 0:
            raise NoReliableDataError('Dataset contains no reliable data.')
        distances = distances[mask]
        acf = acf[mask]
    return distances, acf


# Register analysis functions from this module
NonuniformLineScanInterface.register_function('autocorrelation_from_profile', height_difference_autocorrelation)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import io

###

CHANNEL_NAME_INFO_KEY = 'channel_name'


###

def is_binary_stream(fobj):
    """Check whether fobj is a binary stream"""
    return isinstance(fobj, io.BytesIO) or (
            hasattr(fobj, 'mode') and 'b' in fobj.mode)


def text(func):
    """Decorator that turns the first argument into a binary stream"""

    def func_wrapper(fobj, *args, **kwargs):
        close_file = False
        if not hasattr(fobj, 'read'):
            fobj = open(fobj, 'r', encoding='utf-8')
            fobj_text = fobj
            close_file = True
        elif is_binary_stream(fobj):
            fobj_text = io.TextIOWrapper(fobj, encoding='utf-8')
        else:
            fobj_text = fobj

        try:
            retvals = func(fobj_text, *args, **kwargs)
        finally:
            if is_binary_stream(fobj):
                fobj_text.detach()
                fobj_text = fobj
            if close_file:
                fobj_text.close()
        return retvals

    return func_wrapper


class OpenFromAny(object):
    """
    Context manager for turning file names or already open streams into a
    single stream format for subsequent reading. The file is left in an
    identical state when the context manager returns.
    """

    def __init__(self, fobj, mode='r'):
        """
        Open file

        Arguments
        ---------
        fobj : str or stream
            The file to be opened, specified either as a file name or a
            stream object.
        mode : str
            Open as text ('r') or binary ('rb').
        """
        self._fobj = fobj
        self._mode = mode
        self._fstream = None
        self._stream_position = None
        self._already_open = None

    def __enter__(self):
        # depending from where this function is called, self._fobj might already
        # be a filestream
        self._already_open = False
        self._prior_position = None
        if not hasattr(self._fobj, 'read'):
            # This is a string
            self._fstream = open(self._fobj, self._mode)
        else:
            # This is a stream that is already open
            self._already_open = True
            if hasattr(self._fobj, 'tell'):
                self._prior_position = self._fobj.tell()
            if self._mode == 'rb':
                # Turn this into a binary stream, if it is a text stream
                if isinstance(self._fobj, io.TextIOBase):
                    # file was opened without the 'b' option, so read its buffer to get the binary data
                    self._fstream = self._fobj.buffer
                else:
                    self._fstream = self._fobj
            elif self._mode == 'r':
                if isinstance(self._fobj, io.TextIOBase):
                    # file was opened without the 'b' option, so read its buffer to get the binary data
                    self._fstream = self._fobj
                else:
                    raise ValueError("Don't know how turn a binary stream into a text stream")
            else:
                return ValueError(f"Unknown file open mode '{self._mode}'.")
        self._stream_position = self._fstream.tell()
        return self._fstream

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # Close file, unless it was already open when this object was constructed
        if self._already_open:
            if self._prior_position is not None:
                self._fobj.seek(self._prior_position)
        else:
            self._fstream.close()

        # Reset internal state
        self._fstream = None
        self._stream_position = None
        self._already_open = None



import sys

try:
    import netCDF4
except ImportError:
    netCDF4 = None

mpi_required = sys.argv[1] in ["yes", "1"]

if netCDF4 is not None:
    print('`netCDF4` configuration:')
    print('CDF5 {}'.format(netCDF4.__has_cdf5_format__))
    print('Parallel4: {}'.format(netCDF4.__has_parallel4_support__))
    print('PnetCDF: {}'.format(netCDF4.__has_pnetcdf_support__))

    # Make sure that we have CDF5 support
    assert netCDF4.__has_cdf5_format__

    if mpi_required:
        # Make sure that we have the parallel version running
        #assert netCDF4.__has_parallel4_support__
        assert netCDF4.__has_pnetcdf_support__
else:
    print('`netCDF4` is not available, defaulting to `scipy.io`.')


#
# Copyright 2018-2021 Lars Pastewka
#           2019 Antoine Sanner
#           2019 Michael Röttger
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Power-spectral density for uniform topographies.
"""

import numpy as np

from ..Exceptions import NoReliableDataError, UndefinedDataError
from ..HeightContainer import UniformTopographyInterface
from ..Support import doi
from ..Support.Regression import resample, resample_radial


@doi('10.1088/2051-672X/aa51f8')
def power_spectrum_from_profile(self, window=None, reliable=True, resampling_method='bin-average',
                                collocation='log', nb_points=None, nb_points_per_decade=10):
    """
    Compute power spectrum from 1D FFT(s) of a topography or line scan
    stored on a uniform grid.

    Parameters
    ----------
    self : SurfaceTopography or UniformLineScan
        Container with height information.
    window : str, optional
        Window for eliminating edge effect. See scipy.signal.get_window.
        (Default: no window for periodic Topographies, "hann" window for
        nonperiodic Topographies)
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    resampling_method : str, optional
        Method can be None for no resampling (return on the grid of the
        data) 'bin-average' for simple bin averaging and 'gaussian-process'
        for Gaussian process regression.
        (Default: 'bin-average')
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)

    Returns
    -------
    q : array_like
        Reciprocal space vectors.
    C_all : array_like
        Power spectrum. (Units: length**3)
    """
    if self.has_undefined_data:
        raise UndefinedDataError('This topography has undefined data (missing data points). Power-spectrum cannot be '
                                 'computed for topographies with missing data points.')

    try:
        nx, ny = self.nb_grid_pts
        sx, sy = self.physical_sizes
    except ValueError:
        nx, = self.nb_grid_pts
        sx, = self.physical_sizes

    h = self.window(window).heights()

    # Compute FFT and normalize
    fourier_topography = sx / nx * np.fft.fft(h, axis=0)
    dq = 2 * np.pi / sx
    q = dq * np.arange(nx // 2)

    # This is the raw power spectral density
    C_raw = (np.abs(fourier_topography) ** 2) / sx

    # Fold +q and -q branches. Note: Entry q=0 appears just once, hence
    # exclude from average!
    C_all = C_raw[:nx // 2, ...]
    C_all[1:nx // 2, ...] += C_raw[nx - 1:(nx + 1) // 2:-1, ...]
    C_all /= 2

    if resampling_method is None:
        # Only keep reliable data
        short_cutoff = self.short_reliability_cutoff(2 * sx / nx) if reliable else None
        if short_cutoff is not None:
            mask = q < 2 * np.pi / short_cutoff
            if mask.sum() <= 1:  # There is always q=0 in there
                raise NoReliableDataError('Dataset contains no reliable data.')
            q = q[mask]
            C_all = C_all[mask]
        if self.dim == 1:
            return q, C_all
        else:
            return q, C_all.mean(axis=1)
    else:
        short_cutoff = 2 * np.pi / self.short_reliability_cutoff(2 * sx / nx) if reliable else np.pi * nx / sx
        if collocation == 'log':
            # Exclude zero distance because that does not work on a log-scale
            q = q[1:]
            C_all = C_all[1:]
        if self.dim == 2:
            q = np.resize(q, (C_all.shape[1], q.shape[0])).T.ravel()
            C_all = np.ravel(C_all)
        q, _, C, _ = resample(q, C_all, min_value=q[1], max_value=short_cutoff, collocation=collocation,
                              nb_points=nb_points, nb_points_per_decade=nb_points_per_decade, method=resampling_method)
        return q, C


@doi('10.1088/2051-672X/aa51f8')
def power_spectrum_from_area(self, window=None, reliable=True, collocation='log', nb_points=None,
                             nb_points_per_decade=10, return_map=False, resampling_method='bin-average'):
    """
    Compute power spectrum from 2D FFT and radial average of a topography
    stored on a uniform grid.

    Parameters
    ----------
    self : :obj:`SurfaceTopography`
        Container storing the (two-dimensional) topography map.
    window : str, optional
        Window for eliminating edge effect. See scipy.signal.get_window.
        (Default: None)
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)
    return_map : bool, optional
        Return full 2D power spectrum map. (Default: False)
    resampling_method : str, optional
        Method can be 'bin-average' for simple bin averaging and
        'gaussian-process' for Gaussian process regression.
        (Default: 'bin-average')

    Returns
    -------
    q : array_like
        Reciprocal space vectors.
    C_all : array_like
        Power spectrum. (Units: length**4)
    """
    if self.has_undefined_data:
        raise UndefinedDataError('This topography has undefined data (missing data points). Power-spectrum cannot be '
                                 'computed for topographies with missing data points.')

    nx, ny = self.nb_grid_pts
    sx, sy = self.physical_sizes

    qmax = 2 * np.pi * nx / (2 * sx)

    if reliable:
        # Update qmax
        short_cutoff = self.short_reliability_cutoff()
        if short_cutoff is not None:
            qmax = 2 * np.pi / short_cutoff

    h = self.window(window=window, direction='radial').heights()

    # Compute FFT and normalize
    surface_qk = self.area_per_pt * np.fft.fft2(h)
    C_qk = abs(surface_qk) ** 2 / (sx * sy)  # pylint: disable=invalid-name

    # Radial average
    q_val, q_edges, C_val, _ = resample_radial(C_qk, physical_sizes=(2 * np.pi * nx / sx, 2 * np.pi * ny / sy),
                                               collocation=collocation, nb_points=nb_points,
                                               nb_points_per_decade=nb_points_per_decade, max_radius=qmax,
                                               method=resampling_method)

    if return_map:
        return q_val, C_val, C_qk
    else:
        return q_val, C_val


# Register analysis functions from this module
UniformTopographyInterface.register_function('power_spectrum_from_profile', power_spectrum_from_profile)
UniformTopographyInterface.register_function('power_spectrum_from_area', power_spectrum_from_area)



#
# Copyright 2019-2021 Lars Pastewka
#           2021 Michael Röttger
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from scipy.io import loadmat, whosmat

from ..UniformLineScanAndTopography import Topography
from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo


class MatReader(ReaderBase):
    _format = 'mat'
    _name = 'MATLAB'
    _description = '''
Imports topography data stored in MATLAB workspace files. The reader
automatically extracts all 2D arrays stored in the file and interprets those
as height information. Matlab files do not store units or physical sizes.
These need to be manually provided by the user.
    '''

    def __init__(self, fobj):
        """
        Reads a surface profile from a Matlab file and presents in in a
        SurfaceTopography-conformant manner.

        All two-dimensional arrays present in the matlab data file are
        returned.

        Parameters
        ----------
        fobj : filename or file object
             File to read.
        """
        self._fobj = fobj
        with OpenFromAny(self._fobj, 'rb') as f:
            header = whosmat(f)  # Only read header
            self._channels = []
            for name, shape, data_class in header:
                is_2d_array = False
                try:
                    nx, ny = shape
                    is_2d_array = True
                except (AttributeError, ValueError):
                    pass
                if is_2d_array:
                    channel_info = ChannelInfo(self,
                                               len(self._channels),
                                               name=name,
                                               dim=len(shape),
                                               uniform=True,
                                               nb_grid_pts=shape)
                    # no height scale factor given in mat file

                    self._channels.append(channel_info)

    @property
    def channels(self):
        return self._channels

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={}, periodic=False,
                   subdomain_locations=None, nb_subdomain_grid_pts=None):
        if channel_index is None:
            channel_index = self._default_channel_index

        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError(
                'This reader does not support MPI parallelization.')

        name = self.channels[channel_index].name

        info = info.copy()
        info['data_source'] = name

        with OpenFromAny(self._fobj, 'rb') as f:
            height_data = loadmat(f, variable_names=[name])

        topography = Topography(
            height_data[name], physical_sizes=self._check_physical_sizes(physical_sizes), unit=unit,
            info=info, periodic=periodic)

        if height_scale_factor is not None:
            topography = topography.scale(height_scale_factor)

        return topography

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Filter pipelines for data imputation (filling undefined data points)
"""

import numpy as np
import scipy.sparse

from .GeometryAnalysis import assign_patch_numbers_area, assign_patch_numbers_profile, outer_perimeter_area, \
    outer_perimeter_profile
from ..HeightContainer import UniformTopographyInterface
from ..UniformLineScanAndTopography import DecoratedUniformTopography


class InterpolateUndefinedDataHarmonic(DecoratedUniformTopography):
    """
    Replace undefined data points by interpolation of neighboring
    points with harmonic functions (solutions of the Laplace equation).
    """

    name = 'interpolate_undefined_data_harmonic'

    def __init__(self, topography, info={}):
        super().__init__(topography, info=info)

    @property
    def has_undefined_data(self):
        """
        By definition, this topography has no undefined data.
        """
        return False

    def heights(self):
        """
        Computes the topography with filled in data points.
        """
        heights = self.parent_topography.heights().copy()
        if super().has_undefined_data:
            dim = self.dim

            # Coordinates for each point on the topography
            if dim == 1:
                nx, = self.nb_grid_pts
                x = np.arange(nx)
            elif dim == 2:
                nx, ny = self.nb_grid_pts
                x, y = np.mgrid[:nx, :ny]
            else:
                # Should not happen
                raise NotImplementedError

            # Get undefined data points and identify continuous patches
            mask = np.ma.getmaskarray(heights)
            if dim == 2:
                nb_patches, patch_ids = assign_patch_numbers_area(mask, self.is_periodic)
            else:
                nb_patches, patch_ids = assign_patch_numbers_profile(mask, self.is_periodic)
            assert np.max(patch_ids) == nb_patches

            # We now fill in the patches individually
            for id in range(1, nb_patches + 1):
                # Create mask identifying points in patch
                patch_mask = patch_ids == id
                nb_patch = np.sum(patch_mask)

                # Create mask identifying perimeter points
                if dim == 2:
                    perimeter_mask = outer_perimeter_area(patch_mask, self.is_periodic)
                else:
                    perimeter_mask = outer_perimeter_profile(patch_mask, self.is_periodic)
                nb_perimeter = np.sum(perimeter_mask)

                # Total number of pixels
                nb_pixels = nb_patch + nb_perimeter

                # Create unique pixel indices in patch and perimeter
                pixel_index = np.zeros_like(patch_ids)
                pixel_index[patch_mask] = np.arange(nb_patch)
                pixel_index[perimeter_mask] = np.arange(nb_patch, nb_pixels)

                # Assemble Laplace matrix; diagonal terms
                i0 = np.arange(nb_patch)
                j0 = np.arange(nb_patch)

                # Off-diagonal terms
                i1 = pixel_index[patch_mask]
                j1 = np.roll(pixel_index, 1, 0)[patch_mask]
                i2 = pixel_index[patch_mask]
                j2 = np.roll(pixel_index, -1, 0)[patch_mask]

                if dim == 2:
                    i3 = pixel_index[patch_mask]
                    j3 = np.roll(pixel_index, 1, 1)[patch_mask]
                    i4 = pixel_index[patch_mask]
                    j4 = np.roll(pixel_index, -1, 1)[patch_mask]

                    # Laplace matrix from coordinates
                    laplace = scipy.sparse.coo_matrix(
                        (np.concatenate((-4 * np.ones(nb_patch), np.ones(nb_patch), np.ones(nb_patch),
                                         np.ones(nb_patch), np.ones(nb_patch), np.ones(nb_perimeter))),
                         (np.concatenate((i0, i1, i2, i3, i4, np.arange(nb_patch, nb_pixels))),
                          np.concatenate((j0, j1, j2, j3, j4, np.arange(nb_patch, nb_pixels))))),
                        shape=(nb_pixels, nb_pixels))
                else:
                    # Laplace matrix from coordinates
                    laplace = scipy.sparse.coo_matrix(
                        (np.concatenate((-2 * np.ones(nb_patch), np.ones(nb_patch), np.ones(nb_patch),
                                         np.ones(nb_perimeter))),
                         (np.concatenate((i0, i1, i2, np.arange(nb_patch, nb_pixels))),
                          np.concatenate((j0, j1, j2, np.arange(nb_patch, nb_pixels))))),
                        shape=(nb_pixels, nb_pixels))

                # Dirichlet boundary conditions (heights on perimeter)
                rhs = np.zeros(nb_pixels)
                rhs[nb_patch:] = heights[perimeter_mask]

                # Solve for undefined heights
                heights[patch_mask] = scipy.sparse.linalg.spsolve(laplace.tocsr(), rhs)[:nb_patch]
        return heights


def interpolate_undefined_data(self, method='harmonic'):
    """
    Imputation of undefined data points in topography information that
    typically occurs in optical measurements.

    Parameters
    ----------
    self : SurfaceTopography.Topography or SurfaceTopography.UniformLineScan
        Input topography containing undefined data points.
    method : str
        Imputation methods. Options
           'harmonic': Interpolate with harmonic functions
        (Default: 'harmonic')
    """
    if method == 'harmonic':
        return self.interpolate_undefined_data_with_harmonic_function()
    else:
        raise ValueError(f"Unsupported imputation method '{method}'.")


UniformTopographyInterface.register_function('interpolate_undefined_data',
                                             interpolate_undefined_data)
UniformTopographyInterface.register_function('interpolate_undefined_data_with_harmonic_function',
                                             InterpolateUndefinedDataHarmonic)



#
# Copyright 2018-2021 Lars Pastewka
#           2019-2021 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Bin for small common helper function and classes for uniform
topographies.
"""

import numpy as np

from ..FFTTricks import make_fft
from ..HeightContainer import UniformTopographyInterface
from ..UniformLineScanAndTopography import Topography
from ..UniformLineScanAndTopography import DecoratedUniformTopography


def bandwidth(self):
    """
    Computes lower and upper bound of bandwidth, i.e. of the wavelengths or
    length scales occurring on a topography. The lower end of the bandwidth is
    given by the pixel size, the upper end by the physical dimension. For
    topographies with an aspect ratio that is not unity, this function returns
    the mean value of the two Cartesian directions.

    Returns
    -------
    lower_bound : float
        Lower bound of the bandwidth.
    upper_bound : float
        Upper bound of the bandwidth.
    """
    lower_bound = np.mean(self.pixel_size)
    upper_bound = np.mean(self.physical_sizes)

    return lower_bound, upper_bound


def domain_decompose(self, subdomain_locations, nb_subdomain_grid_pts,
                     communicator=None):
    """
    Turn a topography that is defined over the whole domain into one that is
    decomposed for each individual MPI process.

    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.
    subdomain_locations : tuple of ints
        Origin (location) of the subdomain handled by the present MPI
        process.
    nb_subdomain_grid_pts : tuple of ints
        Number of grid points within the subdomain handled by the present
        MPI process. This is only required if decomposition is set to
        'domain'.
    communicator, optional : mpi4py communicator or NuMPI stub communicator
        Communicator object. Use communicator from topography object if not
        present. (Default: None)

    Returns
    -------
    decomposed_topography : array
        SurfaceTopography object that now holds only data local the MPI
        process.
    """
    if communicator is None and self.communicator is None:
        raise RuntimeError('Please provide an MPI communicator.')
    return self.__class__(self.heights(),
                          self.physical_sizes,
                          periodic=self.is_periodic,
                          decomposition='domain',
                          subdomain_locations=subdomain_locations,
                          nb_subdomain_grid_pts=nb_subdomain_grid_pts,
                          communicator=self.communicator if communicator is None else communicator,
                          unit=self.unit,
                          info=self.info)


def plot(topography, subplot_location=111):
    """
    Plot an image of the topography using matplotlib.

    Parameters
    ----------
    topography : :obj:`SurfaceTopography`
        Height information
    """
    # We import here because we don't want a global dependence on matplotlib
    import matplotlib.pyplot as plt

    try:
        sx, sy = topography.physical_sizes
    except TypeError:
        sx, sy = topography.nb_grid_pts
    nx, ny = topography.nb_grid_pts

    ax = plt.subplot(subplot_location, aspect=1)
    Y, X = np.meshgrid(np.arange(ny + 1) * sy / ny,
                       np.arange(nx + 1) * sx / nx)
    Z = topography[...]
    mesh = ax.pcolormesh(X, Y, Z)
    plt.colorbar(mesh, ax=ax)
    ax.set_xlim(0, sx)
    ax.set_ylim(0, sy)
    if topography.unit is not None:
        unit = topography.unit
    else:
        unit = 'a.u.'
    ax.set_xlabel('Position $x$ ({})'.format(unit))
    ax.set_ylabel('Position $y$ ({})'.format(unit))
    return ax


class FilledTopography(DecoratedUniformTopography):
    def __init__(self, topography, fill_value=-np.infty, info={}):
        """
        masked (undefined) data is replaced with `fill_value`.

        Parameters
        ----------
        topography: Topography or UniformLineScan instance
        fill_value: float or array of floats
            masked value in topography will be replaced
        """
        super().__init__(topography, info=info)
        self.fill_value = fill_value

    def heights(self):
        return np.ma.filled(self.parent_topography.heights(),
                            fill_value=self.fill_value)

    @property
    def has_undefined_data(self):
        return False


# Register analysis functions from this module
UniformTopographyInterface.register_function('make_fft', make_fft)
UniformTopographyInterface.register_function('bandwidth', bandwidth)
UniformTopographyInterface.register_function('domain_decompose', domain_decompose)
Topography.register_function('plot', plot)
UniformTopographyInterface.register_function('fill_undefined_data', FilledTopography)



#
# Copyright 2018-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Power-spectral density for nonuniform topographies.
"""

import numpy as np

from ..Exceptions import ReentrantDataError, NoReliableDataError
from ..HeightContainer import NonuniformLineScanInterface


def sinc(x):
    """sinc function"""
    return np.sinc(x / np.pi)


def dsinc(x):
    """Derivative of the sinc function"""
    tol = 1e-6
    x = np.asarray(x)
    small_values = np.abs(x) < tol
    if small_values.sum() > 0:
        ret = np.zeros_like(x)

        # For small values, use the Taylor series expansion
        # (accurate to O(x^^))
        ret[small_values] = -x[small_values] / 3 + x[small_values] ** 3 / 30

        # For large values, use the normal expression
        large_values = np.logical_not(small_values)
        ret[large_values] = (np.cos(x[large_values]) -
                             sinc(x[large_values])) / x[large_values]
        return ret
    else:
        return (np.cos(x) - sinc(x)) / x


def ft_rectangle(q):
    r"""
    Fourier transform of a rectangle, :math:`f(x) = 1` for :math:`|x| < 1/2`,

    .. math ::

        \tilde{f}(q) = 2\sin(q/2)/q = sinc(q/2)
    """
    # np.sinc is sin(pi*x)/(pi*x), sinc is sin(x)/xCo
    return sinc(q / 2)


def ft_one_sided_triangle(q):
    r"""
    Fourier transform of the one-sided triangle, :math:`f(x) = x` for
    :math:`|x| < 1/2`,

    .. math ::

        \tilde{f}(q) = i [\cos(q/2)/q - 2\sin(q/2)/q^2] = d sinc(q/2)/dq

    Returned value does not contain the factor of :math:`i`, i.e. this
    function is real-valued.
    """
    return dsinc(q / 2) / 2


def apply_window(x, y, window=None):
    if window == 'hann':
        length = x.max() - x.min()
        return (2 / 3) ** (1 / 2) * (1 - np.cos(2 * np.pi * x / length)) * y
    elif window is None or window == 'None':
        return y
    else:
        raise ValueError('Unknown window {}'.format(window))


def power_spectrum(self, reliable=True, algorithm='fft', wavevectors=None, nb_interpolate=5, short_cutoff=np.mean,
                   window=None, resampling_method='bin-average', collocation='log', nb_points=None,
                   nb_points_per_decade=10):
    r"""
    Compute power-spectral density (PSD) for a nonuniform topography. The
    topography is assumed to be given by a series of points connected by
    straight lines.

    Parameters
    ----------
    self : :obj:`NonuniformLineScan`
        Container storing the nonuniform line scan.
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    algorithm : str
        Algorithm to compute autocorrelation.
        * 'fft': Interpolates the nonuniform line scan on a grid and then uses
        the FFT to compute the autocorrelation. Scales O(N log N)
        * 'brute-force': Brute-force computation using between line segements.
        Scale O(N^2 M) where M is the number of distance point.
        (Default: 'fft')
    wavevectors : array, optional
        Wavevectors at which to compute the PSD. If omitted, wavevectors are
        equally spaced with a spacing that corresponds to :math:`2\pi/\lambda`
        where :math:`\lambda` is the shortest distance between two points in
        the `x`-array. (Default: None)
    nb_interpolate : int, optional
        Number of grid points to put between closest points on surface. Only
        used for 'fft' algorithm. (Default: 5)
    short_cutoff : function, optional
        Function that determines how the short cutoff for the returned PSD is
        computed. If set to None, the full PSD of the interpolated data is
        returned. If the user passes a function, that function is called with
        an array that contains the distances between the points of the uniform
        line scan. Pass `np.max`, `np.mean` or `np.min` to use the maximum,
        mean or minimum of that distance as the cutoff.
    window : str, optional
        Name of the window function to apply before computing the PSD.
        Presently only supports Hann window ('hann') or no window (None or
        'None').
        Default: no window for periodic Topographies, "hann" window for
        nonperiodic Topographies
    resampling_method : str, optional
        Method can be None for no resampling (return on the grid of the
        data) 'bin-average' for simple bin averaging and 'gaussian-process'
        for Gaussian process regression.
        (Default: None)
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)

    Returns
    -------
    wavevectors : array
        Wavevector array at which the PSD has been computed. (Unit: length)
    psd : array
        PSD values. (Unit: length**-3)
    """
    if not self.is_periodic and window is None:
        window = "hann"

    s, = self.physical_sizes
    x, y = self.positions_and_heights()
    if algorithm == 'fft':
        if wavevectors is not None:
            raise ValueError("`wavevectors` can only be used with 'brute-force' algorithm.")
        if self.is_reentrant:
            raise ReentrantDataError('This topography is reentrant (i.e. it contains overhangs). The power-spectral '
                                     'density cannot be computed for reentrant topographies.')
        wavevectors, psd = self.to_uniform(nb_interpolate=nb_interpolate) \
            .power_spectrum_from_profile(window=window, reliable=reliable, resampling_method=resampling_method,
                                         collocation=collocation, nb_points=nb_points,
                                         nb_points_per_decade=nb_points_per_decade)
    elif algorithm == 'brute-force':
        if reliable:
            raise ValueError("'Brute-force' algorithm for nonuniform line scans does not support reliability analysis.")
        if resampling_method is not None:
            raise ValueError("'Brute-force' algorithm for nonuniform line scans does not support resampling.")
        if wavevectors is None:
            raise ValueError("You need to specify `wavevectors` for the 'brute-force' algorithm.")

        y = apply_window(x, y, window=window)
        L = x[-1] - x[0]
        if wavevectors is None:
            wavevectors = 2 * np.pi * np.arange(int(L / np.diff(x).min())) / L
        y_q = np.zeros_like(wavevectors, dtype=np.complex128)
        for x1, x2, y1, y2 in zip(x[:-1], x[1:], y[:-1], y[1:]):
            dx = x2 - x1
            if dx > 0:
                dy = y2 - y1
                x0 = (x1 + x2) / 2
                y0 = (y1 + y2) / 2
                y_q += dx * (y0 * ft_rectangle(wavevectors * dx) + 1j * dy *
                             ft_one_sided_triangle(wavevectors * dx)) * np.exp(
                    -1j * x0 * wavevectors)
            else:
                raise ValueError('Nonuniform data points must be sorted in order of ascending x-values.')
        psd = np.abs(y_q) ** 2 / L
    else:
        raise ValueError("Unknown algorithm '{}' specified.".format(algorithm))

    if short_cutoff is not None:
        mask = wavevectors < 2 * np.pi / short_cutoff(np.diff(x))
        if mask.sum() == 0:
            raise NoReliableDataError('Dataset contains no reliable data.')
        wavevectors = wavevectors[mask]
        psd = psd[mask]
    return wavevectors, psd


# Register analysis functions from this module
NonuniformLineScanInterface.register_function('power_spectrum_from_profile', power_spectrum)



#
# Copyright 2020 Lars Pastewka
#           2020 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
useful symbolic expressions for self-affine rough surfaces
"""

from sympy import Symbol, pprint, solve_poly_system, Matrix, zeros
import sympy


q = Symbol('q', positive=True)
q0 = Symbol('q0', positive=True)
q1 = Symbol('q1', positive=True)
h = Symbol('H', positive=True)
alpha = Symbol('α', positive=True)
beta = Symbol('β', positive=True)
pi = Symbol('π', positive=True)
h_rms = Symbol('h_rms', positive=True)

C = beta**2*q**(-2*(h+1))
print("C(q):")
pprint(C)
print()
expr = 1/(2*pi)*q*C
print("integrand")
pprint(expr)
print()

print('q*C')
pprint(sympy.simplify(C*q))
print()
h2 = sympy.simplify(sympy.integrate(expr, q))#(q, q0, q1)))
print("(h_rms)**2:")
pprint(h2)
print()

h2 = 2*pi*sympy.integrate(q*C, (q, q0, q1))
print("(h_rms)**2:")
pprint(h2)
print()

#beta_sol = sympy.solve(h2-h_rms**2, beta)[0]
#print("β:")
#pprint(beta_sol)
#print()

beta_lim = h2.subs({q1: sympy.oo})
beta_lim_sol = sympy.solve(beta_lim-h_rms**2, beta)[0]
print("for large {}, β:".format(q1))
pprint(beta_lim_sol)
print()




#
# Copyright 2020-2021 Lars Pastewka
#           2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

from ..HeightContainer import UniformTopographyInterface
from SurfaceTopography.Support.Interpolation import Bicubic
from ..UniformLineScanAndTopography import Topography
from ..UniformLineScanAndTopography import DecoratedUniformTopography


def interpolate_linear(self):
    r"""
    Returns a linear interpolation function based on the topography's heights.
    For (2D) topographies, the interpolation is not unique. Each rectangle is
    split into two triangles that side in the lower left and upper right part
    of the rectangle.

    Note that for nonperiodic line scans/topographies, interpolation only
    works for positions between 0 and sx - sx/nx - i.e. the interpolated
    physical domain of the topography is smaller than the nominal domain.
    """

    def linear_interpolator_line_scan(x, periodic=None):
        """Override periodicity of underlying topography with the `periodic` argument"""
        scaled_x = x / px
        periodic = is_periodic if periodic is None else periodic
        if not periodic:
            if np.any(scaled_x < 0) or np.any(scaled_x > nx-1):
                raise ValueError('Cannot interpolate outside of physical domain for nonperiodic line scans.')
        int_x = np.array(scaled_x, dtype=int)
        frac_x = scaled_x - int_x
        return (1 - frac_x) * heights[int_x % nx] + frac_x * heights[(int_x + 1) % nx]

    def linear_interpolator_topography(x, y, periodic=None):
        """Override periodicity of underlying topography with the `periodic` argument"""
        scaled_x = x / px
        scaled_y = y / py
        periodic = is_periodic if periodic is None else periodic
        if not periodic:
            if np.any(scaled_x < 0) or np.any(scaled_x > nx-1) or np.any(scaled_y < 0) or np.any(scaled_y > ny-1):
                raise ValueError('Cannot interpolate outside of physical domain for nonperiodic topographies.')
        int_x = np.array(scaled_x, dtype=int)
        int_y = np.array(scaled_y, dtype=int)
        frac_x = scaled_x - int_x
        frac_y = scaled_y - int_y

        int_x %= nx
        int_y %= ny
        int1_x = (int_x + 1) % nx
        int1_y = (int_y + 1) % ny

        lower_triangle = \
            (1 - frac_x - frac_y) * heights[int_x, int_y] + \
            frac_x * heights[int1_x, int_y] + \
            frac_y * heights[int_x, int1_y]
        upper_triangle = \
            (frac_x + frac_y - 1) * heights[int1_x, int1_y] + \
            (1 - frac_x) * heights[int_x, int1_y] + \
            (1 - frac_y) * heights[int1_x, int_y]

        return np.where(frac_x + frac_y > 1, upper_triangle, lower_triangle)

    heights = self.heights()
    is_periodic = self.is_periodic
    if self.dim == 1:
        nx, = self.nb_grid_pts
        px, = self.pixel_size
        return linear_interpolator_line_scan
    elif self.dim == 2:
        nx, ny = self.nb_grid_pts
        px, py = self.pixel_size
        return linear_interpolator_topography
    else:
        raise ValueError(f'Cannot interpolate {self.dim}-dimensional height containers.')


def interpolate_bicubic(self, derivative='fourier'):
    r"""
    Returns a bicubic interpolation function based on the topography's heights
    and slopes as obtained using finite-differences (derivative='fd') or Fourier
    derivative (derivative='fourier').

    Note that for nonperiodic line scans/topographies, interpolation only
    works for positions between 0 and sx - sx/nx - i.e. the interpolated
    physical domain of the topography is smaller than the nominal domain.
    """
    if not self.is_periodic:
        raise ValueError('Bicubic interpolation is only implemented for periodic surfaces.')

    if self.dim != 2:
        raise ValueError('Bicubic interpolation is only implemented for topographies (not line scans).')

    dx, dy = self.pixel_size
    if derivative == 'fourier':
        derx, dery = self.fourier_derivative()
        interp = Bicubic(self.heights(),
                         derx * dx,  # Bicubic assumes a grid spacing of 1
                         dery * dy
                         )
    elif derivative == 'fd':
        # The Bicubic class automatically determines derivatives from fd if left unspecified
        interp = Bicubic(self.heights())
    else:
        raise ValueError(f"Unknown `derivative` type '{derivative}'.")

    def bicubic_interpolator_topography(x, y, derivative=0):
        r"""

        Parameters
        ----------
        x:  array
            coordinate where to interpolate the topography
        y:  array
            coordinate where to interpolate the topography
        derivative: int 0, 1 or 2, optional
            if > 0 returns also derivatives. Note that the 2nd order
            derivatives are not continuous and hence not really usefull

        Returns
        -------
        array of interpolated values or tuple with arrays of values and
        derivatives

        """
        if derivative == 0:
            interp_field = interp(x / dx, y / dy, derivative=derivative)
            return interp_field
        elif derivative == 1:
            interp_field, interp_derx, interp_dery = \
                interp(x / dx, y / dy, derivative=derivative)
            # back to our original grid spacing
            interp_derx = interp_derx / dx
            interp_dery = interp_dery / dy
            return interp_field, interp_derx, interp_dery
        elif derivative == 2:
            interp_field, interp_derx, interp_dery, interp_derxx, interp_deryy, interp_derxy = \
                interp(x / dx, y / dy, derivative=derivative)

            interp_derx = interp_derx / dx
            interp_dery = interp_dery / dy

            interp_derxx = interp_derxx / dx ** 2
            interp_deryy = interp_deryy / dy ** 2
            interp_derxy = interp_derxx / dx / dy

            return interp_field, interp_derx, interp_dery, interp_derxx, interp_deryy, interp_derxy

    return bicubic_interpolator_topography


def interpolate_fourier(self, nb_grid_pts):
    r"""
    Interpolates the heights of `topography` on a finer grid given by
    `nb_grid_pts` by padding the fourier spectrum with zeros.

    Only 2D implemented.

    Note: at the nyquist frequency we use the "minimal-oscillation"
    interpolation, i.e. we assume a cosine wave
    (https://math.mit.edu/~stevenj/fft-deriv.pdf).

    Parameters
    ----------
    self: Topography
    nb_grid_pts: tuple

    Returns
    -------
    SurfaceTopography with interolated values

    """
    bigspectrum = np.zeros((nb_grid_pts[0], nb_grid_pts[1] // 2 + 1),
                           dtype=complex)
    smallspectrum = np.fft.rfft2(self.heights())
    snx, sny = smallspectrum.shape
    nx, ny = self.nb_grid_pts

    # the entries at the nyquist frequency are the superposition of the
    # positive and negative frequency. When we increase the fourier domain, we
    # will split that again between positive and negative, therefore we divide
    # the value by two and copy it to the corresponding negative vector
    if ny % 2 == 0:
        # the nyquist frequency also contains the symmetric and will be
        # twice too big when the symmetric will be included
        smallspectrum[:, -1] = smallspectrum[:, -1] / 2

    if nx % 2 == 0:
        # the nyquist frequency also contains the symmetric and will be
        # twice too big when the symmetric will be included
        smallspectrum[int(nx / 2), :] = smallspectrum[int(nx / 2), :] / 2

    i = snx // 2 if snx % 2 == 0 else (snx - 1) // 2 + 1
    bigspectrum[:i, :sny] = smallspectrum[:i, :sny]
    bigspectrum[-(snx - i):, :sny] = smallspectrum[-(snx - i):, :sny]

    # in the y direction, the entries at the nyquist frequency are
    # automatically symmetrised by the rfft, corresponding to a cosinusoidal
    # wave, i.e. the "minimal-oscillation" trigonometric interpolation
    # (https://math.mit.edu/~stevenj/fft-deriv.pdf).
    # To ensure the same behaviour in the x direction
    # (see test_fourier_interpolate_transpose_symmetry`)`
    # we have to mirror the entries at the niquist frequency by hand
    if snx % 2 == 0:
        bigspectrum[i, :sny] = smallspectrum[i, :sny]

    return Topography(np.fft.irfft2(bigspectrum, s=nb_grid_pts)
                      * np.prod(nb_grid_pts) / np.prod(self.nb_grid_pts),
                      # normalization
                      physical_sizes=self.physical_sizes, periodic=True)


class MirrorStichedTopography(DecoratedUniformTopography):
    """
    adds mirror images at the top (big y) and at the right (big x) of the
    topography, making it kind of periodic

    """

    def __init__(self, parent_topography, info={}):
        if parent_topography.communicator.Get_size() > 1:
            raise (NotImplementedError("MirrorStichedTopography "
                                       "not domain decomposable"))
        super().__init__(parent_topography, info)

    @property
    def is_periodic(self):
        return True

    @property
    def physical_sizes(self):
        return [2 * s for s in self.parent_topography.physical_sizes]

    @property
    def nb_grid_pts(self):
        return [2 * s for s in self.parent_topography.nb_grid_pts]

    def heights(self):
        h = self.parent_topography.heights()
        return np.block([[h[:, :], h[:, ::-1]],
                         [h[:: -1, :], h[::-1, :: -1]]])

    def positions(self):
        nx, ny = self.nb_grid_pts
        lnx, lny = self.nb_subdomain_grid_pts
        sx, sy = self.physical_sizes
        return np.meshgrid(
            (self.subdomain_locations[0] + np.arange(lnx)) * sx / nx,
            (self.subdomain_locations[1] + np.arange(lny)) * sy / ny,
            indexing='ij')


Topography.register_function("mirror_stitch", MirrorStichedTopography)
Topography.register_function("interpolate_linear", interpolate_linear)
Topography.register_function("interpolate_bicubic", interpolate_bicubic)
UniformTopographyInterface.register_function('interpolate_fourier', interpolate_fourier)



#
# Copyright 2020-2021 Lars Pastewka
#           2020-2021 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
from scipy.signal import get_window

from ..FFTTricks import get_window_2D
from ..HeightContainer import UniformTopographyInterface
from ..UniformLineScanAndTopography import DecoratedUniformTopography


class WindowedUniformTopography(DecoratedUniformTopography):
    """
    Construct a topography with a window function applied to it.
    """

    name = 'windowed_topography'

    def __init__(self, topography, window=None, direction=None, info={}):
        """
        window : str, optional
            Window for eliminating edge effect. See scipy.signal.get_window.
            (Default: no window for periodic Topographies, "hann" window for
            nonperiodic Topographies)
        direction : str, optional
            Direction in which the window is applied. Possible options are
            'x', 'y' and 'radial'. If set to None, it chooses 'x' for line
            scans and 'radial' for topographies. (Default: None)
        """
        super().__init__(topography, info=info)

        self._window_name = window
        self._direction = direction

        self._window_data = None

    def _make_window(self):
        self._window_data = None

        n = self.parent_topography.nb_grid_pts

        try:
            nx, ny = n
        except ValueError:
            nx, = n

        window_name = self._window_name
        if not self.parent_topography.is_periodic and window_name is None:
            window_name = "hann"

        direction = self._direction
        if direction is None:
            direction = 'x' if self.parent_topography.dim == 1 else 'radial'

        # Construct window
        if window_name is not None and window_name != 'None':
            if direction == 'x':
                # Get window from scipy.signal
                win = get_window(window_name, nx)
                # Normalize window
                win *= np.sqrt(nx / (win ** 2).sum())
            elif direction == 'y':
                if self.parent_topography.dim == 1:
                    raise ValueError("Direction 'y' does not make sense for line scans.")
                # Get window from scipy.signal
                win = get_window(window_name, ny)
                # Normalize window
                win *= np.sqrt(ny / (win ** 2).sum())
            elif direction == 'radial':
                if self.parent_topography.dim == 1:
                    raise ValueError("Direction 'radial' does not make sense for line scans.")
                win = get_window_2D(window_name, nx, ny,
                                    self.parent_topography.physical_sizes)
                # Normalize window
                win *= np.sqrt(nx * ny / (win ** 2).sum())
            else:
                raise ValueError(f"Unknown direction '{self._direction}'.")

            self._window_data = win

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), \
            self._window_name, self._direction
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._window_name, self._direction = state
        super().__setstate__(superstate)

    @property
    def window_data(self):
        if self._window_data is None:
            self._make_window()
        return self._window_data

    def heights(self):
        """ Computes the windowed topography.
        """
        if self.window_data is None:
            return self.parent_topography.heights()
        else:
            direction = self._direction
            if direction is None:
                direction = 'x' if self.parent_topography.dim == 1 else 'radial'
            if direction == 'x':
                return (self.window_data * self.parent_topography.heights().T).T
            elif direction == 'y' or direction == 'radial':
                return self.window_data * self.parent_topography.heights()
            else:
                raise ValueError(f"Unknown direction '{self._direction}'.")


class FourierFilteredUniformTopography(DecoratedUniformTopography):
    name = 'filtered_topography'

    def __init__(self, topography,
                 filter_function=lambda qx, qy: (np.abs(qx) <= 1) * np.abs(qy) <= 1,
                 isotropic=True,
                 info={}):

        if not topography.is_periodic:
            raise ValueError("only implemented for periodic topographies")
        super().__init__(topography, info=info)

        self._filter_function = filter_function
        self._is_filter_isotropic = isotropic
        # TODO: should be deductible from the filter function signature

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), \
            self._filter_function, self._is_filter_isotropic
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._filter_function, self._is_filter_isotropic = state
        super().__setstate__(superstate)

    @property
    def is_filter_isotropic(self):
        return self._is_filter_isotropic

    def filter_function(self, *args):
        """

        Parameters
        ----------
        if dim = 2 and filter is not isotropic
            qx, qy
        if dim = 1
            q
        """

        if self.dim == 2 and not self.is_filter_isotropic \
                and len(args) != 2:
            raise ("ValueError: qx, qy expected")
        elif self.dim == 1 and len(args) != 1:
            raise ("ValueError: q expected")

        return self._filter_function(*args)

    def heights(self):
        if self.dim == 2:
            nx, ny = self.parent_topography.nb_grid_pts
            sx, sy = self.parent_topography.physical_sizes

            qx = np.arange(0, nx, dtype=np.float64).reshape(-1, 1)
            qx = np.where(qx <= nx // 2, qx / sx,  (qx - nx) / sx)
            qx *= 2 * np.pi

            qy = np.arange(0, ny // 2 + 1, dtype=np.float64).reshape(1, -1)
            qy *= 2 * np.pi / sy

            if self.is_filter_isotropic:
                h_qs = np.fft.irfftn(np.fft.rfftn(self.parent_topography.heights()) *
                                     self.filter_function(np.sqrt(qx ** 2 + qy ** 2)))
            else:
                h_qs = np.fft.irfftn(np.fft.rfftn(self.parent_topography.heights()) *
                                     self.filter_function(qx, qy))

            return h_qs
        elif self.dim == 1:
            s, = self.parent_topography.physical_sizes
            n, = self.parent_topography.nb_grid_pts
            q = abs(2 * np.pi * np.fft.rfftfreq(n, s / n))

            h = self.parent_topography.heights()
            h_q = np.fft.rfft(h)
            h_q_filtered = np.fft.irfft(h_q * self.filter_function(q))

            # Max_imaginary = np.max(np.imag(shifted_pot))
            # assert Max_imaginary < 1e-14 *np.max(np.real(shifted_pot)) ,
            # f"{Max_imaginary}"

            return np.real(h_q_filtered)


class ShortCutTopography(FourierFilteredUniformTopography):
    name = 'shortcut_filtered_topography'

    def __init__(self, topography,
                 cutoff_wavevector=None, cutoff_wavelength=None,
                 kind="circular step",
                 info={}):
        r"""Applies a short wavelength cut filter to the topography using fft.

        for `kind=="circular step"` (default), parts of the spectrum with
        `|q| > cutoff_wavevector` are set to zero

        for `kind=="square step"`, parts of the spectrum with
        `q_x > cutoff_wavevector or q_y > cutoff_wavevector ` are set to zero

        either `cutoff_wavelength` or
        `cutoff_wavevector` :math:`= 2 pi /` `cutoff_wavelength`
        have to be provided.

        Parameters
        ----------
        topography: Topography
        cutoff_wavevector: float
        highest wavevector
        cutoff_wavelength: float
        shortest wavelength
        kind: {"circular step", "square step"}

        Returns
        -------
        Topography with filtered heights

        Examples
        --------
        >>> topography.shortcut(cutoff_wavevector=2 * np.pi / l)
        >>> topography.shortcut(cutoff_wavelength=l) # equivalent

        """
        if not topography.is_periodic:
            raise ValueError("only implemented for periodic topographies")

        if cutoff_wavelength is None:
            if cutoff_wavevector is not None:
                cutoff_wavelength = 2 * np.pi / cutoff_wavevector
            else:
                raise ValueError("cutoff_wavevector "
                                 "or cutoff_wavelength should be provided")
        elif cutoff_wavevector is not None:
            raise ValueError("cutoff_wavevector "
                             "or cutoff_wavelength should be provided")

        self._cutoff_wavelength = cutoff_wavelength
        self._kind = kind

        def circular_step(q):
            return q <= self.cutoff_wavevector

        def square_step(qx, qy):
            return (np.abs(qx) <= self.cutoff_wavevector) * (
                    np.abs(qy) <= self.cutoff_wavevector)

        if self._kind == "circular step":
            super().__init__(topography, info=info,
                             filter_function=circular_step)
        elif self._kind == "square step":
            super().__init__(topography, info=info,
                             filter_function=square_step, isotropic=False)
        else:
            raise ValueError("Invalid kind")

    @property
    def cutoff_wavevector(self):
        return 2 * np.pi / self._cutoff_wavelength

    @property
    def cutoff_wavelength(self):
        return self._cutoff_wavelength

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self._filter_function, \
            self._kind, self._cutoff_wavelength
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._filter_function, self._kind, \
            self._cutoff_wavelength = state
        super().__setstate__(superstate)


class LongCutTopography(FourierFilteredUniformTopography):
    name = 'longcut_filtered_topography'

    def __init__(self, topography,
                 cutoff_wavevector=None, cutoff_wavelength=None,
                 kind="circular step",
                 info={}):
        r"""Applies a long wavelength cut filter to the topography using fft.

        for `kind=="circular step"` (default), parts of the spectrum with
        `|q| < cutoff_wavevector` are set to zero

        for `kind=="square step"`, parts of the spectrum with
        `q_x < cutoff_wavevector or q_y < cutoff_wavevector ` are set to zero

        either `cutoff_wavelength` or
        `cutoff_wavevector` :math:`= 2 pi /` `cutoff_wavelength`
        have to be provided.

        Parameters
        ----------
        topography: Topography
        cutoff_wavevector: float
        highest wavevector
        cutoff_wavelength: float
        shortest wavelength
        kind: {"circular step", "square step"}

        Returns
        -------
        Topography with filtered heights

        Examples
        --------
        >>> topography.longcut(cutoff_wavevector=2 * np.pi / l)
        >>> topography.longcut(cutoff_wavelength=l) # equivalent

        """
        if not topography.is_periodic:
            raise ValueError("only implemented for periodic topographies")

        if cutoff_wavelength is None:
            if cutoff_wavevector is not None:
                cutoff_wavelength = 2 * np.pi / cutoff_wavevector
            else:
                raise ValueError("cutoff_wavevector "
                                 "or cutoff_wavelength should be provided")
        elif cutoff_wavevector is not None:
            raise ValueError("cutoff_wavevector "
                             "or cutoff_wavelength should be provided")

        self._cutoff_wavelength = cutoff_wavelength
        self._kind = kind

        def circular_step(q):
            return q >= self.cutoff_wavevector

        def square_step(qx, qy):
            return (np.abs(qx) >= self.cutoff_wavevector) * (
                    np.abs(qy) >= self.cutoff_wavevector)

        if self._kind == "circular step":
            super().__init__(topography, info=info,
                             filter_function=circular_step)
        elif self._kind == "square step":
            super().__init__(topography, info=info,
                             filter_function=square_step, isotropic=False)
        else:
            raise ValueError("Invalid kind")

    @property
    def cutoff_wavevector(self):
        return 2 * np.pi / self._cutoff_wavelength

    @property
    def cutoff_wavelength(self):
        return self._cutoff_wavelength

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self._filter_function, \
            self._kind, self._cutoff_wavelength
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._filter_function, self._kind, \
            self._cutoff_wavelength = state
        super().__setstate__(superstate)


UniformTopographyInterface.register_function("window", WindowedUniformTopography)
UniformTopographyInterface.register_function("filter", FourierFilteredUniformTopography)
UniformTopographyInterface.register_function("shortcut", ShortCutTopography)
UniformTopographyInterface.register_function("longcut", LongCutTopography)



#
# Copyright 2019-2021 Lars Pastewka
#           2019-2021 Antoine Sanner
#           2019 Michael Röttger
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Support for uniform topogography descriptions
"""

import numpy as np

from NuMPI import MPI
from NuMPI.Tools import Reduction

from .HeightContainer import (
    AbstractTopography,
    UniformTopographyInterface, DecoratedTopography
)
from .Uniform.Detrending import tilt_from_height, tilt_and_curvature
from .Support.UnitConversion import get_unit_conversion_factor


class UniformLineScan(AbstractTopography, UniformTopographyInterface):
    """
    Line scan that lives on a uniform one-dimensional grid.
    """

    def __init__(self, heights, physical_sizes, periodic=False, unit=None, info={}, communicator=MPI.COMM_SELF):
        """
        Parameters
        ----------
        profile : array_like
            Data containing the height information. Needs to be a
            one-dimensional array.
        physical_sizes : tuple of floats
            Physical physical_sizes of the topography map
        periodic : bool, optional
            Flag setting the periodicity of the surface
        unit : str, optional
            The length unit.
        info : dict, optional
            The info dictionary containing auxiliary data.
        """
        if not hasattr(heights, 'ndim'):
            heights = np.array(heights)

        if heights.ndim != 1:
            raise ValueError('Heights array must be one-dimensional.')

        super().__init__(unit=unit, info=info, communicator=communicator)

        # Automatically turn this into a masked array if it is not already a
        # masked array and there is data missing
        if not np.ma.is_masked(heights) and np.sum(np.logical_not(np.isfinite(heights))) > 0:
            heights = np.ma.masked_where(np.logical_not(np.isfinite(heights)), heights)
        self._heights = heights
        self._size = np.asarray(physical_sizes).item()
        self._periodic = periodic

    def __getstate__(self):
        state = super().__getstate__(), self._heights, self._size, self._periodic
        return state

    def __setstate__(self, state):
        superstate, self._heights, self._size, self._periodic = state
        super().__setstate__(superstate)

    # Implement abstract methods of AbstractHeightContainer

    @property
    def dim(self):
        return 1

    @property
    def physical_sizes(self):
        if self._size is None:
            return None
        return self._size,

    @property
    def is_periodic(self):
        return self._periodic

    # Implement uniform line scan interface

    @property
    def nb_grid_pts(self):
        return len(self._heights),

    @property
    def nb_subdomain_grid_pts(self):
        return self.nb_grid_pts

    @property
    def is_domain_decomposed(self):
        return False

    @property
    def pixel_size(self):
        return (self.physical_sizes[0] / self.nb_grid_pts[0],)

    @property
    def area_per_pt(self):
        return self.pixel_size[0]

    @property
    def has_undefined_data(self):
        return np.ma.getmaskarray(self._heights).sum() > 0

    def positions(self, meshgrid=True):
        """
        Return grid positions.

        Arguments
        ---------
        meshgrid : bool, optional
            Ignore, for compatibility with two-dimensional topographies

        Returns
        -------
        x : np.ndarray
            X-positions
        """
        r, = self.nb_grid_pts
        p, = self.pixel_size
        return np.arange(r) * p

    def heights(self):
        return self._heights

    def save(self, fname, compress=True):
        """
        saves the topography as a NumpyTxtTopography. Warning: This only saves
        the profile; the physical_sizes is not contained in the file
        """
        if compress:
            if not fname.endswith('.gz'):
                fname = fname + ".gz"
        np.savetxt(fname, self.heights())

    def squeeze(self):
        return self


class Topography(AbstractTopography, UniformTopographyInterface):
    """
    Topography that lives on a uniform two-dimensional grid, i.e. a
    topography map.
    """

    def __init__(self, heights, physical_sizes, periodic=False,
                 nb_grid_pts=None, subdomain_locations=None,
                 nb_subdomain_grid_pts=None,
                 unit=None, info={},
                 decomposition='serial',
                 communicator=MPI.COMM_SELF):

        """
        Parameters
        ----------
        heights : array_like
            Data containing the height information. Needs to be a
            two-dimensional array.
        physical_sizes : tuple of floats
            Physical physical_sizes of the topography map
        periodic : bool, optional
            The topography is periodic. (Default: False)
        nb_grid_pts : tuple of ints, optional
            Number of grid points for the full topography. This is only
            required if decomposition is set to 'subdomain'.
        subdomain_locations : tuple of ints, optional
            Origin (location) of the subdomain handled by the present MPI
            process.
        nb_subdomain_grid_pts : tuple of ints, optional
            Number of grid points within the subdomain handled by the present
            MPI process. This is only required if decomposition is set to
            'domain'.
        unit : str, optional
            The length unit.
        info : dict, optional
            The info dictionary containing auxiliary data.
        decomposition : str, optional
            Specification of the data decomposition of the heights array. If
            set to 'subdomain', the heights array contains only the part of
            the full topography local to the present MPI process. If set to
            'domain', the heights array contains the global array.
            Default: 'serial', which fails for parallel runs.
        communicator : mpi4py communicator or NuMPI stub communicator, optional
            The MPI communicator object.
            default value is COMM_SELF because sometimes NON-MPI readers that
            do not set the communicator value are used in MPI Programs.
            See discussion in issue #166

        Examples for initialisation:
        ----------------------------
        1. Serial code
        >>>Topography(heights, physical_sizes, [periodic, info])
        2. Parallel code, providing local data
        >>>Topography(heights, physical_sizes, [periodic],
                      decomposition='subdomain',
                      subdomain_locations, nb_grid_pts, comm,
                      [info])
        3. Parallel code, providing full data
        >>>Topography(heights, physical_sizes, [periodic],
                      decomposition='domain',
                      subdomain_locations, nb_subdomain_grid_pts, comm,
                      [info])
        """
        if heights.ndim != 2:
            raise ValueError('Heights array must be two-dimensional.')

        super().__init__(unit=unit, info=info, communicator=communicator)

        # Automatically turn this into a masked array if there is data missing
        if not np.ma.is_masked(heights) and np.sum(np.logical_not(np.isfinite(heights))) > 0:
            heights = np.ma.masked_where(np.logical_not(np.isfinite(heights)), heights)

        if communicator.Get_size() == 1:
            # case 1. : no parallelization
            if nb_grid_pts is not None and tuple(nb_grid_pts) != heights.shape:
                raise ValueError(
                    'This is a serial run but `nb_grid_pts` (= {}) does not '
                    'equal the shape of the `heights` (= {}) array.'.format(nb_grid_pts, heights.shape))
            if subdomain_locations is not None and tuple(
                    subdomain_locations) != (0, 0):
                raise ValueError(
                    'This is a serial run but `subdomain_locations` (= {}) '
                    'is not the origin.'.format(subdomain_locations))
            if nb_subdomain_grid_pts is not None and tuple(
                    nb_subdomain_grid_pts) != heights.shape:
                raise ValueError(
                    'This is a serial run but `nb_subdomain_grid_pts` '
                    '(= {}) does not equal the shape '
                    'of the `heights` (= {}) array.'.format(nb_subdomain_grid_pts, heights.shape))
            self._nb_grid_pts = heights.shape
            self._subdomain_locations = (0, 0)
            self._heights = np.asanyarray(heights)
        elif decomposition == 'subdomain':
            # Case 2.: parallelized and local data provided
            if nb_grid_pts is None:
                raise ValueError(
                    "This is a parallel run with 'subdomain' decomposition; '"
                    "'please specify `nb_grid_pts` since it cannot be "
                    "inferred.")
            if subdomain_locations is None:
                raise ValueError('This is a parallel run; please specify '
                                 '`subdomain_locations`.')
            if nb_subdomain_grid_pts is not None and tuple(
                    nb_subdomain_grid_pts) != heights.shape:
                raise ValueError(
                    "This is a parallel run with 'subdomain' decomposition "
                    "but `nb_subdomain_grid_pts` (= {}) does not equal the "
                    "shape of the `heights` (= {}) array.".format(nb_subdomain_grid_pts, heights.shape))
            self._nb_grid_pts = nb_grid_pts
            self._subdomain_locations = subdomain_locations
            self._heights = np.asanyarray(heights)
        elif decomposition == 'domain':
            # Case 3: parallelized but global data provided
            if nb_grid_pts is not None and tuple(nb_grid_pts) != heights.shape:
                raise ValueError(
                    "This is a parallel run with 'domain' decomposition but "
                    "`nb_grid_pts` (= {}) does not equal the shape of the "
                    "`heights` (= {}) array.".format(nb_grid_pts, heights.shape))
            if subdomain_locations is None:
                raise ValueError('This is a parallel run; please specify `subdomain_locations`.')
            if nb_subdomain_grid_pts is None:
                raise ValueError(
                    "This is a parallel run with 'domain' decomposition; "
                    "please specify `nb_subdomain_grid_pts` since it cannot "
                    "be inferred.")
            self._nb_grid_pts = heights.shape
            self._subdomain_locations = subdomain_locations
            self._heights = np.asanyarray(
                heights[tuple(slice(s, s + n) for s, n in
                              zip(subdomain_locations,
                                  nb_subdomain_grid_pts))])
        elif decomposition == 'serial':
            raise ValueError(
                "`decomposition` is 'serial' but this is a parallel run.")
        else:
            raise ValueError(
                "`decomposition` can be either 'domain' or 'subdomain'.")

        self._size = physical_sizes
        self._periodic = periodic

    def __getstate__(self):
        state = super().__getstate__(), self._heights, self._size, self._periodic, self._subdomain_locations, \
                self._nb_grid_pts
        return state

    def __setstate__(self, state):
        superstate, self._heights, self._size, self._periodic, self._subdomain_locations, self._nb_grid_pts = state
        super().__setstate__(superstate)

    # Implement abstract methods of AbstractHeightContainer

    @property
    def dim(self):
        return 2

    @property
    def is_periodic(self):
        return self._periodic

    @property
    def physical_sizes(self):
        return self._size

    @property
    def nb_grid_pts(self):
        return self._nb_grid_pts

    @property
    def nb_subdomain_grid_pts(self):
        """ needs to be testable to make sure that geometry and halfspace are
            compatible
        """
        return self._heights.shape

    @property
    def subdomain_locations(self):
        return self._subdomain_locations

    @property
    def subdomain_slices(self):
        return tuple(slice(s, s + n) for s, n in
                     zip(self.subdomain_locations, self.nb_subdomain_grid_pts))

    # Implement topography interface

    @property
    def pixel_size(self):
        return np.asarray(self.physical_sizes) / np.asarray(self.nb_grid_pts)

    @property
    def area_per_pt(self):
        return np.prod(self.pixel_size)

    @property
    def has_undefined_data(self):
        return Reduction(self._communicator).sum(np.ma.getmaskarray(self._heights).sum()) > 0

    def positions(self, meshgrid=True):
        """
        Return grid positions.

        Arguments
        ---------
        meshgrid : bool, optional
            If True, return the position on the same grid as the heights.
            Otherwise, return one-dimensional position arrays. (Default: True)

        Returns
        -------
        x : np.ndarray
            X-positions
        y : np.ndarray
            Y-positions
        """
        # FIXME: Write test for this method
        nx, ny = self.nb_grid_pts
        lnx, lny = self.nb_subdomain_grid_pts
        sx, sy = self.physical_sizes
        x = (self.subdomain_locations[0] + np.arange(lnx)) * sx / nx
        y = (self.subdomain_locations[1] + np.arange(lny)) * sy / ny
        if meshgrid:
            x, y = np.meshgrid(x, y, indexing='ij')
        return x, y

    def heights(self):
        return self._heights

    def positions_and_heights(self, **kwargs):
        x, y = self.positions(**kwargs)
        return x, y, self.heights()

    @property
    def communicator(self):
        return self._communicator

    def squeeze(self):
        return self


class DecoratedUniformTopography(DecoratedTopography,
                                 UniformTopographyInterface):
    @property
    def has_undefined_data(self):
        return self.parent_topography.has_undefined_data

    @property
    def is_domain_decomposed(self):
        return False

    @property
    def is_periodic(self):
        return self.parent_topography.is_periodic

    @property
    def dim(self):
        return self.parent_topography.dim

    @property
    def pixel_size(self):
        return self.parent_topography.pixel_size

    @property
    def unit(self):
        if self._unit is None:
            return self.parent_topography.unit
        else:
            return self._unit

    @property
    def info(self):
        info = self.parent_topography.info
        info.update(self._info)
        if self.unit is not None:
            info.update(dict(unit=self.unit))
        return info

    @property
    def physical_sizes(self):
        return self.parent_topography.physical_sizes

    @property
    def nb_grid_pts(self):
        return self.parent_topography.nb_grid_pts

    @property
    def nb_subdomain_grid_pts(self):
        """ needs to be testable to make sure that geometry and halfspace are
            compatible
        """
        return self.parent_topography.nb_subdomain_grid_pts

    @property
    def subdomain_locations(self):
        return self.parent_topography.subdomain_locations

    @property
    def subdomain_slices(self):
        return self.parent_topography.subdomain_slices

    @property
    def area_per_pt(self):
        return self.parent_topography.area_per_pt

    def positions(self, **kwargs):
        return self.parent_topography.positions(**kwargs)

    def positions_and_heights(self, **kwargs):
        if self.dim == 1:
            return self.positions(**kwargs), self.heights()
        else:
            return (*self.positions(**kwargs), self.heights())

    def squeeze(self):
        if self.dim == 1:
            return UniformLineScan(self.heights(), self.physical_sizes,
                                   periodic=self.is_periodic, info=self.info)
        else:
            return Topography(self.heights(), self.physical_sizes,
                              periodic=self.is_periodic, info=self.info)


class ScaledUniformTopography(DecoratedUniformTopography):
    """Scale heights, positions or both."""

    def __init__(self, topography, unit=None, info={}):
        """
        This topography wraps a parent topography and rescales x, y and z
        coordinates according to certain rules.

        Arguments
        ---------
        topography : :obj:`UniformTopographyInterface`
            Parent topography
        unit : str
            Target unit.
        info : dict, optional
            Updated entries to the info dictionary. (Default: {})
        """
        super().__init__(topography, unit=unit, info=info)

    @property
    def height_scale_factor(self):
        return get_unit_conversion_factor(self.parent_topography.unit, self.unit)

    @property
    def position_scale_factor(self):
        return get_unit_conversion_factor(self.parent_topography.unit, self.unit)

    @property
    def pixel_size(self):
        """Compute rescaled pixel sizes."""
        return tuple(self.position_scale_factor * s for s in super().pixel_size)

    @property
    def physical_sizes(self):
        """Compute rescaled physical sizes."""
        return tuple(self.position_scale_factor * s for s in super().physical_sizes)

    def positions(self, **kwargs):
        """Compute the rescaled positions."""
        if self.dim == 1:
            return self.position_scale_factor * super().positions(**kwargs)
        else:
            return tuple(self.position_scale_factor * p for p in super().positions(**kwargs))

    def heights(self):
        """Computes the rescaled profile."""
        return self.height_scale_factor * self.parent_topography.heights()


class StaticallyScaledUniformTopography(ScaledUniformTopography):
    """Scale heights, positions or both."""

    def __init__(self, topography, height_scale_factor, position_scale_factor=1, unit=None, info={}):
        """
        This topography wraps a parent topography and rescales x, y and z
        coordinates according to certain rules.

        Arguments
        ---------
        topography : :obj:`UniformTopographyInterface`
            Parent topography
        height_scale_factor : float
            Factor to scale heights with.
        position_scale_factor : float, optional
            Factor to scale lateral positions (`physical_sizes`, etc.) with.
            (Default: 1)
        unit : str, optional
            Target unit. This is simply used to update the metadata, not for
            determining scale factors. (Default: None)
        info : dict, optional
            Updated entries to the info dictionary. (Default: {})
        """
        super().__init__(topography, unit=unit, info=info)
        self._height_scale_factor = height_scale_factor
        self._position_scale_factor = position_scale_factor

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self._height_scale_factor, self._position_scale_factor
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._height_scale_factor, self._position_scale_factor = state
        super().__setstate__(superstate)

    @property
    def height_scale_factor(self):
        return self._height_scale_factor

    @property
    def position_scale_factor(self):
        return self._position_scale_factor


class DetrendedUniformTopography(DecoratedUniformTopography):
    """
    Remove trends from a topography. This is achieved by fitting polynomials
    to the topography data to extract trend lines. The resulting topography
    is then detrended by substracting these trend lines.

    Note on periodicity: Detrended Topographies with mode other than 'center'
    will have `is_periodic` property set to False.
    """

    def __init__(self, topography, detrend_mode='height', info={}):
        """
        Note on periodicity: Detrended Topographies with mode other than
        "center" will have `is_periodic` property set to False.

        Parameters
        ----------
        topography : Topography
            SurfaceTopography to be detrended.
        detrend_mode : str
            'center': center the topography, no trend correction.
            'height': adjust slope such that rms height is minimized.
            'slope': adjust slope such that rms slope is minimized.
            'curvature': adjust slope and curvature such that rms height is
            minimized.
            (Default: 'height')
        """
        super().__init__(topography, info=info)
        self._detrend_mode = detrend_mode
        self._detrend()

    def _detrend(self):
        if self.dim == 1:
            if self._detrend_mode == 'center':
                self._coeffs = (self.parent_topography.mean(),)
            elif self._detrend_mode == 'height':
                x, y = self.parent_topography.positions_and_heights()
                a1, a0 = np.polyfit(x / self.parent_topography.physical_sizes, y, 1)
                self._coeffs = a0, a1
            elif self._detrend_mode == 'slope':
                sl = self.parent_topography.derivative(1, periodic=False).mean()
                n, = self.nb_grid_pts
                s, = self.physical_sizes
                grad = sl * s
                self._coeffs = [self.parent_topography.mean() - grad * (n - 1) / (2 * n), grad]
            elif self._detrend_mode == 'curvature':
                x, y = self.parent_topography.positions_and_heights()
                a2, a1, a0 = np.polyfit(
                    x / self.parent_topography.physical_sizes, y, 2)
                self._coeffs = a0, a1, a2
            else:
                raise ValueError("Unsupported detrend mode '{}' for line scans.".format(self._detrend_mode))
        else:  # self.dim == 2
            if self._detrend_mode is None or self._detrend_mode == 'center':
                self._coeffs = [self.parent_topography.mean()]
            elif self._detrend_mode == 'height':
                self._coeffs = [s for s in tilt_from_height(self.parent_topography)]
            elif self._detrend_mode == 'slope':
                slx, sly = self.parent_topography.derivative(1, periodic=False)
                slx = slx.mean()
                sly = sly.mean()
                nx, ny = self.nb_grid_pts
                sx, sy = self.physical_sizes
                self._coeffs = [
                    slx * sx, sly * sy,
                    self.parent_topography.mean() - slx * sx * (nx - 1) / (2 * nx) - sly * sy * (ny - 1) / (2 * ny)
                ]
            elif self._detrend_mode == 'curvature':
                self._coeffs = [s for s in tilt_and_curvature(self.parent_topography)]
            else:
                raise ValueError("Unsupported detrend mode '{}' for 2D topographies.".format(self._detrend_mode))

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self._detrend_mode, self._coeffs
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self._detrend_mode, self._coeffs = state
        super().__setstate__(superstate)

    @property
    def coeffs(self):
        return self._coeffs

    @property
    def detrend_mode(self):
        return self._detrend_mode

    @detrend_mode.setter
    def detrend_mode(self, detrend_mode):
        self._detrend_mode = detrend_mode
        self._detrend()

    @property
    def is_periodic(self):
        """
        SurfaceTopography stays periodic only after detrend mode "center".
        Otherwise the detrended SurfaceTopography is non-periodic.
        """
        if self.detrend_mode == "center":
            return self.parent_topography.is_periodic
        else:
            return False

    def heights(self):
        """ Computes the combined profile.
        """
        if len(self._coeffs) == 1:
            a0, = self._coeffs
            return self.parent_topography.heights() - a0
        elif self.dim == 1:
            x = np.arange(self.nb_grid_pts[0]) / self.nb_grid_pts[0]
            if len(self._coeffs) == 2:
                a0, a1 = self._coeffs
                return self.parent_topography.heights() - a0 - a1 * x
            elif len(self._coeffs) == 3:
                a0, a1, a2 = self._coeffs
                return self.parent_topography.heights() - a0 - a1 * x - a2 * x * x
            else:
                raise RuntimeError('Unknown physical_sizes of coefficients tuple for line scans.')
        else:  # self.dim == 2
            x, y = np.meshgrid(*(np.arange(n) / n for n in self.nb_grid_pts), indexing='ij')
            if len(self._coeffs) == 3:
                a1x, a1y, a0 = self._coeffs
                return self.parent_topography.heights() - a0 - a1x * x - a1y * y
            elif len(self._coeffs) == 6:
                m, n, mm, nn, mn, h0 = self._coeffs
                xx = x * x
                yy = y * y
                xy = x * y
                return self.parent_topography.heights() - h0 - m * x - n * y - mm * xx - nn * yy - mn * xy
            else:
                raise RuntimeError('Unknown physical_sizes of coefficients '
                                   'tuple for 2D topographies.')

    def stringify_plane(self, fmt=lambda x: str(x)):
        """
        Return a string giving the expression for the detrending plane.
        """

        str_coeffs = [fmt(x) for x in self._coeffs]
        if self.dim == 1:
            if len(self._coeffs) == 1:
                h0, = str_coeffs
                return h0
            elif len(self._coeffs) == 2:
                return '{0} + {1} x'.format(*str_coeffs)
            elif len(self._coeffs) == 3:
                return '{0} + {1} x + {2} x^2'.format(*str_coeffs)
            else:
                raise RuntimeError(
                    'Unknown physical_sizes of coefficients tuple.')
        else:
            if len(self._coeffs) == 1:
                h0, = str_coeffs
                return h0
            elif len(self._coeffs) == 3:
                return '{2} + {0} x + {1} y'.format(*str_coeffs)
            elif len(self._coeffs) == 6:
                return '{5} + {0} x + {1} y + {2} x^2 + {3} y^2 + {4} xy' \
                    .format(*str_coeffs)
            else:
                raise RuntimeError(
                    'Unknown physical_sizes of coefficients tuple.')

    @property
    def curvatures(self):
        r"""
        Curvature(s) of the fitted plane.

        Returns
        -------
        :math:`\rho = 1 / R` for line scans or tuple :math:`\rho_{xx}, \rho_{yy}, \rho_{xy}` for topographies
        """

        if self.dim == 1:
            if len(self._coeffs) == 3:
                return (2 * self._coeffs[2]) / self.physical_sizes[0] ** 2,
            elif len(self._coeffs) in {1, 2}:
                return 0,
            else:
                raise RuntimeError(
                    'Unknown physical_sizes of coefficients tuple.')
        else:
            if len(self._coeffs) == 6:
                sx, sy = self.physical_sizes
                return (2 * self._coeffs[2]) / sx ** 2, (
                        2 * self._coeffs[3]) / sy ** 2, (
                               2 * self._coeffs[4]) / (sx * sy)
            elif len(self._coeffs) in {1, 3}:
                return 0, 0, 0
            else:
                raise RuntimeError(
                    'Unknown physical_sizes of coefficients tuple.')


class TransposedUniformTopography(DecoratedUniformTopography):
    """
    Tranpose topography.
    """

    def __init__(self, topography, info={}):
        """
        Parameters
        ----------
        topography : :obj:`UniformTopography`
            SurfaceTopography to transpose
        info : dict
            Additional entries for the info dictionary
        """
        super().__init__(topography, info=info)

    @property
    def nb_grid_pts(self):
        """ Return number of points """
        if self.dim == 1:
            return self.parent_topography.nb_grid_pts
        else:
            nx, ny = self.parent_topography.nb_grid_pts
            return ny, nx

    @property
    def physical_sizes(self):
        """ Return physical physical_sizes """
        if self.dim == 1:
            return self.parent_topography.physical_sizes
        else:
            sx, sy = self.parent_topography.physical_sizes
            return sy, sx

    def heights(self):
        """ Computes the rescaled profile.
        """
        return self.parent_topography.heights().T

    def positions(self, **kwargs):
        X, Y = self.parent_topography.positions(**kwargs)
        return Y.T, X.T


class TranslatedTopography(DecoratedUniformTopography):
    """ used when geometries are translated
    """
    name = 'translated_topography'

    def __init__(self, topography, offset=(0, 0), info={}):
        """
        Keyword Arguments:
        topography  -- SurfaceTopography to translate
        offset -- Translation offset in number of grid points
        """
        super().__init__(topography, info=info)
        assert isinstance(topography, Topography)
        self._offset = offset

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset, offsety=None):
        if offsety is None:
            self.offset = offset
        else:
            self.offset = (offset, offsety)

    def heights(self):
        """ Computes the translated profile.
        """
        offsetx, offsety = self.offset
        return np.roll(
            np.roll(self.parent_topography.heights(), offsetx, axis=0),
            offsety, axis=1)


class CompoundTopography(DecoratedUniformTopography):
    """ used when geometries are combined
    """
    name = 'compound_topography'

    def __init__(self, topography_a, topography_b):
        """ Behaves like a topography that is a sum of two Topographies
        Keyword Arguments:
        topography_a   -- first topography of the compound
        topography_b   -- second topography of the compound
        """
        super().__init__()

        def combined_val(prop_a, prop_b, propname):
            """
            topographies can have a fixed or dynamic, adaptive nb_grid_pts (or
            other attributes). This function assures that -- if this function
            is called for two topographies with fixed nb_grid_ptss --
            the nb_grid_pts are identical
            Parameters:
            prop_a   -- field of one topography
            prop_b   -- field of other topography
            propname -- field identifier (for error messages only)
            """
            if prop_a is None:
                return prop_b
            else:
                if prop_b is not None:
                    assert prop_a == prop_b, \
                        "{} incompatible:{} <-> {}".format(
                            propname, prop_a, prop_b)
                return prop_a

        self._dim = combined_val(topography_a.dim, topography_b.dim, 'dim')
        self._nb_grid_pts = combined_val(topography_a.nb_grid_pts,
                                         topography_b.nb_grid_pts,
                                         'nb_grid_pts')
        self._size = combined_val(topography_a.physical_sizes,
                                  topography_b.physical_sizes,
                                  'physical_sizes')
        self.parent_topography_a = topography_a
        self.parent_topography_b = topography_b

    def heights(self):
        """ Computes the combined profile
        """
        return (self.parent_topography_a.heights() +
                self.parent_topography_b.heights())


# Register analysis functions from this module
UniformTopographyInterface.register_function('mean', lambda this: Reduction(this._communicator).mean(this.heights()))
UniformTopographyInterface.register_function('min', lambda this: Reduction(this._communicator).min(this.heights()))
UniformTopographyInterface.register_function('max', lambda this: Reduction(this._communicator).max(this.heights()))

# Register pipeline functions from this module
UniformTopographyInterface.register_function('to_unit', ScaledUniformTopography)
UniformTopographyInterface.register_function('scale', StaticallyScaledUniformTopography)
UniformTopographyInterface.register_function('detrend', DetrendedUniformTopography)
UniformTopographyInterface.register_function('transpose', TransposedUniformTopography)
UniformTopographyInterface.register_function('translate', TranslatedTopography)



#
# Copyright 2020 Lars Pastewka
#           2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import matplotlib.pyplot as plt
import numpy as np

from SurfaceTopography.Support.Interpolation import Bicubic
from SurfaceTopography.Generation import fourier_synthesis

# from muFFT import FourierInterpolation # future


nx, ny = [512] * 2
sx, sy = [1.] * 2

# %% Generate random topography
hc = 0.1 * sx

topography = fourier_synthesis((nx, ny), (sx, sy), 0.8, rms_height=1.,
                               short_cutoff=hc, long_cutoff=hc + 1e-9, )
topography = topography.scale(1 / topography.rms_height_from_area())
dx, dy = topography.fourier_derivative()

# %%
fig, ax = plt.subplots()
ax.imshow(topography.heights())
fig.show()
# %%
fig, ax = plt.subplots()
ax.imshow(dx)
fig.show()
fig, ax = plt.subplots()
ax.imshow(topography.derivative(1)[0])
fig.show()

# %% check bicubic interpolation against fourier interpolation

# %%
fig, ax = plt.subplots()
x, y = topography.positions()
ax.plot(x[:, 0], topography.heights()[:, 0], ".k", label="original")

skips = [4, 8, 16, 32, 64]
rms_err = []
max_err = []
for skip in skips:
    grid_slice = (slice(None, None, skip), slice(None, None, skip))

    interp = Bicubic(topography.heights()[grid_slice],
                     dx[grid_slice] * topography.pixel_size[0] * skip,
                     dy[grid_slice] * topography.pixel_size[1] * skip
                     )

    interp_field, interp_derx, interp_dery = interp(
        x / (topography.pixel_size[0] * skip),
        y / (topography.pixel_size[1] * skip), derivative=1)
    l, = ax.plot(x[grid_slice][:, 0], topography.heights()[grid_slice][:, 0],
                 "+")
    ax.plot(x[:, 0], interp_field[:, 0], color=l.get_color(),
            label=r"bicubic, $l_{{cor}} / \Delta_x={}$"
            .format(hc / (skip * topography.pixel_size[0])))

    rms_err.append(
        np.sqrt(np.mean((interp_field - topography.heights()) ** 2)))
    max_err.append(np.max(abs(interp_field - topography.heights())))
    ax.legend()
    fig.show()

skips = np.array(skips)
rms_err = np.array(rms_err)
max_err = np.array(max_err)

# %%

fig, ax = plt.subplots()
sampling = (skips * topography.pixel_size[0]) / hc

ax.plot(sampling, rms_err, "-o", label="rms error")
ax.plot(sampling, max_err, "-o", label="max error")

ax.set_xlabel(r"$\Delta_x / l_{cor}$")
ax.legend()
ax.set_yscale("log")
ax.set_xscale("log")
fig.show()
